/*
  $Id: File.java 7015 2015-11-17 12:57:57Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.DirectoryData;
import net.sf.basedb.core.data.FileAttachableData;
import net.sf.basedb.core.data.FileData;
import net.sf.basedb.core.data.FileSetData;
import net.sf.basedb.core.data.ItemSubtypeData;
import net.sf.basedb.core.data.MimeTypeData;
import net.sf.basedb.util.EqualsHelper;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.MD5;
import net.sf.basedb.util.StreamCacher;
import net.sf.basedb.util.uri.ConnectionParameters;
import net.sf.basedb.util.uri.ConnectionManager;
import net.sf.basedb.util.uri.ConnectionManagerUtil;
import net.sf.basedb.util.uri.ResumableConnectionManager;
import net.sf.basedb.util.uri.UriMetadata;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.Charset;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;
import java.security.MessageDigest;

/**
	This class is used to represent individual file items
	and information about them.
	
	This item also handels the physical file on disk. It should
	work as expected, but under very unfortunate circumstances, 
	ie. an error happens during rollback or commit after the 
	changers has been written to the database, it may not delete
	all files that should be deleted.
	
	@author enell
	@version 2.0
	@base.modified $Date: 2015-11-17 13:57:57 +0100 (ti, 17 nov 2015) $
*/
public class File
	extends CommonItem
	implements Transactional, DiskConsumable, Subtypable
{
	/**
		The type of item represented by this class.
		@see Item#FILE ITEM.FILE
		@see #getType() getType()
	*/
	public static final Item TYPE = Item.FILE;

	/**
		The id for the <code>FileType</code> item representing a
		protocol.
		@since 3.0
	*/
	public static final String PROTOCOL = "net.sf.basedb.core.FileType.PROTOCOL";
	
	/**
		The id for the <code>FileType</code> item representing a
		raw data file.
		@since 3.0
	*/
	public static final String RAW_DATA = "net.sf.basedb.core.FileType.RAW_DATA";
	
	/**
		The id for the <code>FileType</code> item representing a
		plate mapping file.
		@since 3.0
	*/
	public static final String PLATE_MAPPING = "net.sf.basedb.core.FileType.PLATE_MAPPING";
	
	/**
		The id for the <code>FileType</code> item representing a
		reporter map file.
		@since 3.0
	*/
	public static final String REPORTER_MAP = "net.sf.basedb.core.FileType.REPORTER_MAP";
	
	/**
		The id for the <code>FileType</code> item representing a
		print map file.
		@since 3.0
	*/
	public static final String PRINT_MAP = "net.sf.basedb.core.FileType.PRINT_MAP";
	
	/**
		The id for the <code>FileType</code> item representing an
		image file.
		@since 3.0
	*/
	public static final String IMAGE = "net.sf.basedb.core.FileType.IMAGE";
	
	/**
		The id for the <code>FileType</code> item representing a
		spot images file.
		@since 3.0
	*/
	public static final String SPOT_IMAGES = "net.sf.basedb.core.FileType.SPOT_IMAGES";
	
	/**
		The id for the <code>FileType</code> item representing a
		reporter file.
		@since 3.0
	*/
	public static final String REPORTER = "net.sf.basedb.core.FileType.REPORTER";
	
	/**
		The id for the <code>FileType</code> item representing a
		plate file.
		@since 3.0
	*/
	public static final String PLATE = "net.sf.basedb.core.FileType.PLATE";
	
	/**
		The id for the <code>FileType</code> item representing a
		file containing analysed spot data.
		@since 3.0
	*/
	public static final String SPOT_DATA = "net.sf.basedb.core.FileType.SPOT_DATA";
	
	/**
		The id for the <code>FileType</code> item representing a
		file containing analysed extra value data.
		@since 3.0
	*/
	public static final String EXTRAVALUE_DATA = "net.sf.basedb.core.FileType.EXTRAVALUE_DATA";

	/**
		The id for the <code>FileType</code> item representing a
		file containing aligned sequence data.
		@since 3.0
	*/
	public static final String ALIGNED_SEQUENCE = "net.sf.basedb.core.FileType.ALIGNED_SEQUENCE";

	
	/**
		Create a new <code>File</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param directory The {@link Directory} that holds the file. It may
			not be null.
		@return The new <code>File</code> item
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public static File getNew(DbControl dc, Directory directory)
		throws BaseException
	{
		File file = dc.newItem(File.class);
		file.setAction(File.Action.NOTHING);
		file.setDirectory(directory);
		file.setLocation(Location.OFFLINE);
		file.setName("New file");
		file.uploadCompressed = directory.getAutoCompress();
		if (directory.getAutoShare() && directory.isShared())
		{
			ShareableUtil.copyPermissions(dc, directory, file);
		}
		return file;
	}

	/**
		Get a <code>File</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>File</code> item
		@throws ItemNotFoundException If an item with the specified id 
			is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static File getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		File fi = dc.loadItem(File.class, id);
		if (fi == null) throw new ItemNotFoundException("File[id=" + id + "]");
		return fi;
	}
	
	/**
		Get a <code>File</code> item when you know the path. Use the create
		parameter to specify if a new file should be create if it doesn't
		exists. The directory part of the path must still be an existing
		path, otherwise an <code>ItemNotFoundException</code> is thrown. 
		
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param path The path to the file to load
		@param create TRUE if a new file should be created if one doesn't
			eixts, FALSE to throw an exception (DbControl.saveItem must always be
			called)
		@return The <code>File</code> item
		@throws ItemNotFoundException If an item with the specified path 
			is not found and create is FALSE
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static File getByPath(DbControl dc, Path path, boolean create)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		int directoryId = 0;
		File file = null;
		try
		{
			directoryId = Directory.getIdFromPath(dc, path);
		}
		catch (ItemNotFoundException ex)
		{
			throw new ItemNotFoundException("File[path="+path+"]");
		}
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), "GET_FILE_IN_DIRECTORY");
		/*
			SELECT f
			FROM File f
			WHERE f.directory = :directory
			AND f.name = :name
		*/
		query.setInteger("directory", directoryId);
		query.setString("name", path.getFilename());
		FileData fileData = HibernateUtil.loadData(FileData.class, query);
		
		if (fileData == null)
		{
			if (!create)
			{
				throw new ItemNotFoundException("File[path="+path+"]");
			}
			else
			{
				Directory d = Directory.getByPath(dc, path);
				file = getNew(dc, d);
				file.setName(path.getFilename());
			}
		}
		else
		{
			file = dc.getItem(File.class, fileData);
		}
		return file;
	}

	/**
		Get a <code>File</code> item when you know the directory and filename.
		Use the create parameter to specify if a new file should be create if 
		it doesn't exists.  
		
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param directory The directory where the file is located.
		@param fileName The name of the file.
		@param create TRUE if a new file should be created if one doesn't
			eixts, FALSE to throw an exception (DbControl.saveItem must always be
			called)
		@return The <code>File</code> item
		@throws ItemNotFoundException If an item with the specified path 
			is not found and create is FALSE
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static File getFile(DbControl dc, Directory directory, String fileName, boolean create)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (fileName != null) fileName = fileName.trim();
		File file = null;
		if (directory == null)
		{
			throw new InvalidUseOfNullException("directory");
		}
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), "GET_FILE_IN_DIRECTORY");
		/*
			SELECT f
			FROM File f
			WHERE f.directory = :directory
			AND f.name = :name
		*/
		query.setInteger("directory", directory.getId());
		query.setString("name", fileName);
		FileData fileData = HibernateUtil.loadData(FileData.class, query);
		
		if (fileData == null)
		{
			if (!create)
			{
				throw new ItemNotFoundException("File[directory="+directory.getPath()+"/"+fileName+"]");
			}
			else
			{
				file = getNew(dc, directory);
				file.setName(fileName);
			}
		}
		else
		{
			file = dc.getItem(File.class, fileData);
		}
		return file;
	}
	
	/**
		Checks if a file with the specified name exists in the directory.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param directory The directory to check in.
		@param fileName The name of the file to look for.
		@return TRUE if the file exists, FALSE otherwise.
	*/
	public static boolean exists(DbControl dc, Directory directory, String fileName)
		throws BaseException
	{
		if (fileName != null) fileName = fileName.trim();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), "GET_FILE_IN_DIRECTORY");
		/*
			SELECT f
			FROM File f
			WHERE f.directory = :directory
			AND f.name = :name
		*/
		query.setInteger("directory", directory.getData().getId());
		query.setString("name", fileName);
		FileData file = HibernateUtil.loadData(FileData.class, query);
		return file != null;
	}
	
	/**
		Get a query configured to retrieve <code>File</code> items.
		
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<File> getQuery()
	{
		return new ItemQuery<File>(File.class);
	}

	/**
		The path to a newly uploaded file.
	*/
	private java.io.File newInternalFile;

	/**
		Files that are EXTERNAL are cached locally per transaction
	*/
	private StreamCacher downloadCache;
	
	/**
		The path to the internal file.
	*/
	private java.io.File internalFile;
	
	/**
		A stream for uploading a physical file.
	*/
	private UploadStream uploadStream;
	
	/**
		If the next upload should use compression or not.
	*/
	private boolean uploadCompressed;
	
	/**
	 	Creates a new file item from the given data.
	 	
		@param data the data.
	*/
	File(FileData data)
	{
		super(data);
	}

	@Override
	FileData getData()
	{
		return (FileData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the Nameable interface
		-------------------------------------------
	*/
	@Override
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		if (!Path.isValidName(name)) throw new InvalidDataException("A file cannot contain the following characters "+Path.INVALID_CHARACTERS);
		super.setName(name);
	}
	// -------------------------------------------
	/*
		From the DiskConsumning interface
		-------------------------------------------
	*/
	/**
		Always FALSE.
	*/
	@Override
	public boolean useDifference()
	{
		return false;
	}
	@Override
	public long getBytes()
	{
		long bytes = getData().getCompressedSize(); 
		return bytes < 0 ? 0 : bytes;
	}
	/**
		Always 0.
	*/
	@Override
	public long getDifference()
	{
		return 0;
	}
	/**
		Always {@link QuotaType#FILE}.
	*/
	@Override
	public String getQuotaTypeSystemId() 
	{
		return QuotaType.FILE;
	}
	@Override
	public Location getLocation()
	{
		return Location.fromValue(getData().getLocation());
	}
	// -------------------------------------------
	/*
		From the Subtypable interface
		-----------------------------
	*/
	@Override
	public ItemSubtype getItemSubtype()
	{
		return getDbControl().getItem(ItemSubtype.class, getData().getItemSubtype());
	}
	@Override
	public void setItemSubtype(ItemSubtype subtype)
	{
		checkPermission(Permission.WRITE);
		if (subtype != null)
		{
			subtype.setOnItem(this);
		}
		else
		{
			getData().setItemSubtype(null);
		}
	}
	// -------------------------------------------
	/*
		From BasicItem
		-------------------------------------------
	*/
	/**
		Checks if:
		<ul>
		<li>A {@link FileAttachable} item is using this key.
		</ul>
		@return TRUE if this item is used, FALSE otherwise
		@throws BaseException If there is another error
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "FIND_USED_FILES");
		/*
			SELECT fad
			FROM net.sf.basedb.core.data.FileAttachableData fad
			WHERE fad.file = :file
		*/
		query.setEntity("file", this.getData());
		boolean used = HibernateUtil.loadData(FileAttachableData.class, query) != null;
		if (!used)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_FILESETMEMBERS_FOR_FILE", "count(*)");
			/*
				SELECT {1}
				FROM FileSetMemberData fsm
				WHERE fsm.file = :file
			*/
			query.setEntity("file", this.getData());
			used = HibernateUtil.loadData(Long.class, query) > 0;
		}
		if (!used)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_SPOTIMAGES_FOR_FILE", "count(*)");
			/*
				SELECT {1}
				FROM SpotImagesData spi
				WHERE spi.redImageFile = :file 
				OR spi.greenImageFile = :file 
				OR spi.blueImageFile = :file 
				OR spi.spotImagesFile = :file 
			*/
			query.setEntity("file", this.getData());
			used = HibernateUtil.loadData(Long.class, query) > 0;
		}
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link FileAttachable} items with this file attached
		<li>{@link SpotImages} using this file
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// FileAttachable items
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"FIND_USED_FILES");
			/*
				SELECT fad
				FROM net.sf.basedb.core.data.FileAttachableData fad
				WHERE fad.file = :file
			*/
		query.setEntity("file", this.getData());
		addUsingItems(using, query);
		
		// It is the FileStoreEnabled item that is the using item, not the FileSet
		// So, we need to do a bit of hard work to make it happen
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_FILESETMEMBERS_FOR_FILE", "fsm.fileSet");
			/*
				SELECT {1}
				FROM FileSetMemberData fsm
				WHERE fsm.file = :file
			*/
		query.setEntity("file", this.getData());
		List<FileSetData> fileSets = HibernateUtil.loadList(FileSetData.class, query, getSessionControl());
		for (FileSetData fs : fileSets)
		{
			Item itemType = Item.fromValue(fs.getItemType());
			String hql = "SELECT item.id"+
				" FROM "+itemType.getDataClass().getName()+" item"+
				" WHERE item.fileSet = :fileSet";
			query = HibernateUtil.createQuery(session, hql);
			query.setEntity("fileSet", fs);
			int itemId = HibernateUtil.loadData(Integer.class, query);
			using.add(new ItemProxy(itemId, itemType));
		}		
		
		// SpotImages
		query = HibernateUtil.getPredefinedQuery(session, "GET_SPOTIMAGES_FOR_FILE", "spi.id");
			/*
				SELECT {1}
				FROM SpotImagesData spi
				WHERE spi.redImageFile = :file 
				OR spi.greenImageFile = :file 
				OR spi.blueImageFile = :file 
				OR spi.spotImagesFile = :file 
			*/
		query.setEntity("file", this.getData());
		addUsingItems(using, Item.SPOTIMAGES, query);

		return using;
	}

	@Override
	void onBeforeCommit(Transactional.Action action)
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (uploadStream != null)
		{
			try
			{
				uploadStream.close();
				uploadStream = null;
			}
			catch (IOException ex)
			{}
		}
		if (getInternalName() != null)
		{
			// an old file exists
			internalFile = getAbsolutePath(getInternalName());
		}
		if (newInternalFile != null)
		{
			// a new file has been uploaded
			getData().setLastUpdate(new Date());
			setInternalName(getRelativePath(newInternalFile));
		}
		if (downloadCache != null)
		{
			downloadCache.close();
			downloadCache = null;
		}
	}

	@Override
	void onAfterCommit(Transactional.Action action)
	{
		super.onAfterCommit(action);
		if ((getLocation() != Location.PRIMARY) || (action == Transactional.Action.DELETE))
		{
			// The file has been moved or deleted from the primary location
			if (internalFile != null) internalFile.delete();
			if (newInternalFile != null) newInternalFile.delete();
		}
		else if (newInternalFile != null)
		{
			// A new file has been uploaded -> replace the existing file
			if ((internalFile != null) && !internalFile.equals(newInternalFile))
			{
				internalFile.delete();
			}
		}
		newInternalFile = null;
		internalFile = null;
	}

	@Override
	void onRollback(Transactional.Action action)
	{
		super.onRollback(action);
		if (uploadStream != null)
		{
			try
			{
				uploadStream.close();
			}
			catch (IOException ex)
			{}
		}
		if (newInternalFile != null)
		{
			newInternalFile.delete();
		}
		if (downloadCache != null)
		{
			downloadCache.close();
			downloadCache = null;
		}
		newInternalFile = null;
		internalFile = null;
	}

	/**
		Check the write protected flag and deny DELETE if it is set.
	*/
	@Override
	void initPermissions(int granted, int denied)
	{
		if (isWriteProtected()) denied |= Permission.deny(Permission.DELETE);
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------
	
	/**
		Get the internal name of this file.
		
		@return The internal name.
	*/
	private String getInternalName()
	{
		return getData().getInternalName();
	}
	
	/**
		Set the internal name of this file.
		
		@param internalName The internal filename.
	*/
	private void setInternalName(String internalName)
	{
		getData().setInternalName(internalName);
	}
	
	/**
		Set the location of this <code>File</code> item.
		@param location The new <code>Location</code>
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission, or if the file is write protected and the new location isn't
			{@link Location#PRIMARY}
		@throws InvalidUseOfNullException If location is null
	*/
	public void setLocation(Location location) 
		throws PermissionDeniedException, InvalidUseOfNullException
	{
		checkPermission(Permission.WRITE);
		if (location == null) throw new InvalidUseOfNullException("location");
		if (isWriteProtected() && location != Location.PRIMARY)
		{
			throw new PermissionDeniedException("The file is write protected: " + this);
		}
		getData().setLocation(location.getValue());
		if (location != Location.EXTERNAL) getData().setUrl(null);
	}

	/**
	 	Get the {@link File.Action} of this file.
	 	@return the action
	*/
	public File.Action getAction()
	{
		return File.Action.fromValue(getData().getAction());
	}

	/**
		Set the action of this <code>File</code> item. This method call 
		has no affect on external files.
		@param action The new <code>File.Action</code>
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidUseOfNullException If action is null
	*/
	public void setAction(File.Action action) 
		throws PermissionDeniedException, InvalidUseOfNullException
	{
		checkPermission(Permission.WRITE);
		if (action == null) throw new InvalidUseOfNullException("action");
		if (getLocation() == Location.EXTERNAL) return;
		getData().setAction(action.getValue());
	}	

	/**
		Get the (uncompressed) size of this file.
		@return The size in bytes or -1 if not known
		@see #getCompressedSize()
	*/
	public long getSize()
	{
		return getData().getSize();
	}
	
	/**
		Get the compressed size of this file. If the file is not stored in a compressed
		format this method return the same value as {@link #getSize()}
		@return The compressed size in bytes or -1 if not known
		@see #isCompressed()
		@since 2.5
	*/
	public long getCompressedSize()
	{
		return getData().getCompressedSize();
	}

	/**
		If this file is stored in a compressed format or not.
		@return TRUE if the file is compressed
		@since 2.5
	*/
	public boolean isCompressed()
	{
		return getData().isCompressed();
	}
	
	/**
		Get the MD5 hash of the file contents, or null if not known.
		@return A 32-character hexadecimal string
	*/
	public String getMd5()
	{
		return getData().getMd5();
	}
	
	/**
		The maximum length of the MIME type that can be stored in the database.
		@see #setMimeType(String)
	*/
	public static final int MAX_MIMETYPE_LENGTH = FileData.MAX_MIMETYPE_LENGTH;
	
	/**
		The mimetype of this <code>File</code>. It can be a defined {@link MimeType}
		, any mimetype or null.
		@return the mimetype of this file
	*/
	public String getMimeType()
	{
		return getData().getMimeType();
	}
	
	/**
		Set the mime type for this <code>File</code> item. It can be any mimetype or one
		avalible in the mimetype table.
		@param mimeType The new mime type for this item
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the mime type is longer than {@link #MAX_MIMETYPE_LENGTH}
	*/
	public void setMimeType(String mimeType)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setMimeType(StringUtil.setNullableString(mimeType, "mimeType", MAX_MIMETYPE_LENGTH));
	}
	
	/**
		The maximum length of the character set that can be stored in the database.
		@see #setCharacterSet(String)
		@since 2.9
	*/
	public static final int MAX_CHARSET_LENGTH = FileData.MAX_CHARSET_LENGTH;
	/**
		Get the character set used in this file (for text-files only).
		@return The name of the character set as returned by
			{@link Charset#name()}, or null if not known or applicable to 
			the file type
	 	@since 2.9
	*/
	public String getCharacterSet()
	{
		return getData().getCharacterSet();
	}
	/**
		Set the character set used in this file (for text-files only).
		@param charset The name of the character set as returned by
			{@link Charset#name()}
		@since 2.9
	*/
	public void setCharacterSet(String charset)
	{
		checkPermission(Permission.WRITE);
		getData().setCharacterSet(StringUtil.setNullableString(charset, "characterSet", MAX_CHARSET_LENGTH));
	}

	
	/**
		Get the associated {@link Directory} item.
		@return The <code>Directory</code> item
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have {@link Permission#READ READ}
			permission to the items
		@throws BaseException If there is another error
	*/
	public Directory getDirectory()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Directory.class, getData().getDirectory());
	}

	/**
		Set the directory of this <code>File</code> item.
		@param directory The new {@link Directory}
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#WRITE} permission to the file or
			{@link Permission#USE} permission to the directory
		@throws InvalidUseOfNullException If directory is null
	*/
	public void setDirectory(Directory directory)
		throws PermissionDeniedException, InvalidUseOfNullException
	{
		checkPermission(Permission.WRITE);
		if (directory == null) throw new InvalidUseOfNullException("directory");
		directory.checkPermission(Permission.USE);
		getData().setDirectory(directory.getData());
	}
	
	/**
		Check if the file data is write protected. A write protected file 
		can't be deleted, moved offline or to the secondary storage, or replaced
		with another file. This setting doesn't affect metadata about the file, 
		such as the name, description, file type, mime type, etc.
		@return TRUE if the file is write protected, FALSE otherwise
		@since 2.4
	*/
	public boolean isWriteProtected()
	{
		return getData().isWriteProtected();
	}
	
	/**
		Set the write protected flag for this file.
		@param writeProtected TRUE if the file should be write protected
	 	@throws PermissionDeniedException If the logged in user doesn't 
	 		have write permission on the file item. 
		@since 2.4
		@see #isWriteProtected()
	*/
	public void setWriteProtected(boolean writeProtected)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setWriteProtected(writeProtected);
	}
	
	/**
		Get the date and time the actual file data was last updated. This
		value is only affected by file uploads, not by changing the name, 
		description or any other metadata.
		@return The date or null if no file has been uploaded
		@since 2.4
	*/
	public Date getLastUpdate()
	{
		return DateUtil.copy(getData().getLastUpdate());
	}
	
	/**
		The maximum length of the URL type that can be stored in the database.
		@see #setUrl(String, boolean)
		@since 2.16
	*/
	public static final int MAX_URL_LENGTH = FileData.MAX_URL_LENGTH;
	
	/**
		Get the URL to the (external) file. This property is only set/valid
		for files that are stored externally (getLocation() == Location.EXTERNAL)
		@return The URL or null if the file is stored internally
		@since 2.16
	*/
	public String getUrl()
	{
		return getLocation() == Location.EXTERNAL ? getData().getUrl() : null;
	}
	
	/**
		Get the URL of an external file as an URI. This property is only set/valid
		for files that are stored externally (getLocation() == Location.EXTERNAL).
		If the URL is is not a valid URI an exception is thrown.
		@return An URI or null
		@since 3.0
	*/
	public URI getURI()
	{
		return getLocation() == Location.EXTERNAL ? URI.create(getData().getUrl()) : null;
	}
	
	/**
		Set the URL to an external file that should represent this file
		object. NOTE! The location, size, md5, internal file, etc. will be reset 
		as a result of this call. NOTE 2! If the file server is also updated 
		at the same time it is important that the file server has been changed
		before setting the URL, otherwise the old file server is used.
		
		@param url The url to the file (null is not allowed)
		@param loadMetadata If TRUE, a HEAD request will be issued to the given
			URL and metadata such as size, MIME type, etc. will be extracted if
			given in the response
		@since 2.16
		@throws InvalidDataException If the URL is null, too long, or not a valid URL
		@throws PermissionDeniedException If the logged in user doesn't have write permission
	*/
	public void setUrl(String url, boolean loadMetadata)
	{
		checkPermission(Permission.WRITE);
		FileData data = getData();
		// If the url is the same we simple return
		if (EqualsHelper.equals(url, data.getUrl()) && !loadMetadata) return;
		
		// Validate the URL
		URI uri = null;
		try
		{
			uri = new URI(url).normalize();
			data.setUrl(StringUtil.setNotNullString(uri.toString(), "url", MAX_URL_LENGTH));
		}
		catch (URISyntaxException ex)
		{
			throw new InvalidDataException("Invalid URL: " + url, ex);
		}
		
		// Update properties so they make sense for external files
		data.setLocation(Location.EXTERNAL.getValue());
		data.setCompressed(false);
		data.setMd5(null);
		data.setAction(Action.NOTHING.getValue());
		
		// Reset metadata (will be updated later if loadMetadata is true)
		data.setSize(-1);
		data.setCompressedSize(-1);
		data.setLastUpdate(new Date());
		if (loadMetadata)
		{
			updateURLMetadata();
		}
		// Use auto-detection by file extension if no MIME type has been set
		if (data.getMimeType() == null)
		{
			setMimeTypeAuto(null, null);
		}
	}

	/**
		If this is an external file, try to update the metadata.
		@since 3.3
	*/
	public void updateURLMetadata()
	{
		if (getLocation() != Location.EXTERNAL) return;
		
		checkPermission(Permission.WRITE);
		
		FileData data = getData();
		data.setSize(-1);
		data.setCompressedSize(-1);
		data.setLastUpdate(new Date());
		
		ConnectionParameters server = getServerInfo();
		ConnectionManager manager = ConnectionManagerUtil.createConnectionManager(this, server);
		try
		{
			setMetadataFromURI(manager.getMetadata());
		}
		catch (IOException ex)
		{
			throw new BaseException("Could not load metadata for URL:" + data.getUrl(), ex);
		}
	}
	
	/**
		Get the file server on which the external file is located. The file server
		can be used to store username/password that is required to access the file.
		External files that doesn't require authentication usually don't need to be
		associated with a server.
		@return A FileServer or null if no file server has been set
		@since 2.16
	*/
	public FileServer getFileServer()
	{
		return getDbControl().getItem(FileServer.class, getData().getFileServer());
	}
	
	/**
		Set the file server on which this external file is located.
		
		@param fileServer The new {@link FileServer}
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#WRITE} permission to the file or
			{@link Permission#USE} permission to the file server
		@since 2.16
	*/
	public void setFileServer(FileServer fileServer)
		throws PermissionDeniedException, InvalidUseOfNullException
	{
		checkPermission(Permission.WRITE);
		if (fileServer != null) fileServer.checkPermission(Permission.USE);
		getData().setFileServer(fileServer == null ? null : fileServer.getData());
	}
	
	private ConnectionParameters getServerInfo()
	{
		return ConnectionParameters.create(getData().getFileServer());
	}
	
	private void setMetadataFromURI(UriMetadata metadata)
	{
		FileData data = getData();
		Long length = metadata.getLength();
		data.setSize(length == null ? -1 : length);
		
		Date lastUpdated = metadata.getLastModified();
		if (lastUpdated != null) data.setLastUpdate(lastUpdated);
		
		String mimeType = metadata.getMimeType();
		if (mimeType != null) data.setMimeType(mimeType);
		
		String charset = metadata.getCharacterSet();
		if (charset != null) data.setCharacterSet(charset);
		
		String md5 = metadata.getMd5();
		if (md5 != null) data.setMd5(md5);
	}
	
	/**
		Get the complete path of this file.
		@return A {@link Path} object
	*/
	public Path getPath()
	{
		DirectoryData dd = getData().getDirectory();
		LinkedList<String> directories = new LinkedList<String>();
		while (dd != null && dd.getId() != SystemItems.getId(Directory.ROOT))
		{
			directories.addFirst(dd.getName());
			dd = dd.getParent();
		}
		return new Path(null, directories.toArray(new String[directories.size()]), getName());
	}

	/**
		Set the mime type automatically by checking the extension of the
		name. If no match is found, the mime type is set to the default value
		if specified, or left unchanged if the default is null. This method will 
		also check the {@link MimeType#getAutoCompress()} flag and select
		auto-compression for the next upload if it is set. Auto-compression
		is also selected if the {@link Directory#getAutoCompress()} flag is
		set.
		
		@param defaultMimeType The default value to use if no match is found for the file
			extension
		@param defaultFileType The default file type to set if no match is found or null
			to leave it unchanged
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#WRITE} permission
		@see #uploadCompressed()
		@since 3.0
	*/
	public void setMimeTypeAuto(String defaultMimeType, ItemSubtype defaultFileType)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		String name = getName();
		int dot = name.lastIndexOf('.');
		if (dot >= 0)
		{
			String extension = name.substring(dot + 1);
			try
			{
				org.hibernate.Session session = getDbControl().getHibernateSession();
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_MIME_TYPE_FROM_EXTENSION");
				/*
					SELECT mt
					FROM mimetype mt
					WHERE mt.extension = :extension
				*/
				query.setString("extension", extension);
				MimeTypeData mimeType = HibernateUtil.loadData(MimeTypeData.class, query);
				ItemSubtypeData fileType = null;
				if (mimeType != null)
				{
					setMimeType(mimeType.getName());
					fileType = mimeType.getFileType();
					uploadCompressed |= mimeType.getAutoCompress() && Application.autoCompressionEnabled();
				}
				else if (defaultMimeType != null)
				{
					setMimeType(defaultMimeType);
				}
				if (fileType != null)
				{
					getData().setItemSubtype(fileType);
				}
				else if (defaultFileType != null)
				{
					setItemSubtype(defaultFileType);
				}
			}
			catch (BaseException e)
			{}
		}
	}
	
	/**
		Return the result of auto-compress check from the
		last call to {@link #setMimeTypeAuto(String, ItemSubtype)}. 
		If this method returns true, the next call to {@link #upload(InputStream, boolean)}
		or {@link #getUploadStream(boolean)} will store the uploaded file in a
		compressed format.
		
		@return TRUE if the next upload should be compressed
		@since 2.5
	*/
	public boolean uploadCompressed()
	{
		return uploadCompressed;
	}
	
	/**
		Compress the file stored on the disk. If the file is already
		compressed or not located in the {@link Location#PRIMARY}
		location this method does nothing.
		@param progress An optional progress reporter
		@throws PermissionDeniedException If the logged in user doesn't have write
			permission on the file
		@since 2.5
	*/
	public void compress(ProgressReporter progress)
	{
		checkPermission(Permission.WRITE);
		if (!isCompressed() && getLocation() == Location.PRIMARY)
		{
			compressOrDecompress(progress, true);
		}
	}
	
	/**
		Decompress the file stored on the disk. If the file isn't 
		compressed or not located in the {@link Location#PRIMARY}
		location this method does nothing.
		@param progress An optional progress reporter
		@throws PermissionDeniedException If the logged in user doesn't have write
			permission on the file
		@since 2.5
	*/
	public void decompress(ProgressReporter progress)
	{
		if (isCompressed() && getLocation() == Location.PRIMARY)
		{
			compressOrDecompress(progress, false);
		}
	}
	
	private void compressOrDecompress(ProgressReporter progress, boolean compress)
	{
		InputStream compressed = getDownloadStream(0);
		OutputStream uncompressed = getUploadStream(true, compress);
		AbsoluteProgressReporter aProgress = null;
		if (progress != null)
		{
			aProgress = new SimpleAbsoluteProgressReporter(progress, getSize());
			String msg = compress ? "Compressing " : "Decompressing ";
			aProgress.displayAbsolute(0, msg + getName());
		}
		try
		{
			FileUtil.copy(compressed, uncompressed, aProgress);
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			try
			{
				uncompressed.close();
				compressed.close();
			}
			catch (IOException ex)
			{}
		}
	}
	
	/**
		Upload a physical file for this file item. This method calls
		{@link #upload(InputStream, boolean, Boolean)}. A null value is used 
		for the compress parameter to enable auto-compression.
	 	@param in Input stream from the file to be uploaded. 
	 	@param checkMd5 TRUE if the Md5 sum should be checked, FALSE otherwise.
	 	@throws PermissionDeniedException If the logged in user doesn't have permission 
	 		to upload the file
	 	@throws BaseException If there is some other error. 
	*/
	public void upload(InputStream in, boolean checkMd5)
		throws PermissionDeniedException, BaseException
	{
		upload(in, checkMd5, null);
	}

	/**
		Upload a physical file for this file item.
		@param in An <code>InputStream</code> which will be read
			until end of file is found
		@param checkMd5 If the MD5 sum of the uploaded data should
			be checked against the MD5 sum already stored in the database.
			If no MD5 exists in the database this parameter is ignored
		@param compress TRUE to store the file in a compressed format, or null
			to use whatever the {@link Directory#getAutoCompress()} is specifying
			or what auto-detection from the {@link #setMimeTypeAuto(String, ItemSubtype)}
			found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#WRITE} permission for this item, or if the
			file is write protected
		@throws BaseException If there is another error
		@since 2.5
	*/
	public void upload(InputStream in, boolean checkMd5, Boolean compress)
		throws PermissionDeniedException, BaseException
	{
		OutputStream out = getUploadStream(checkMd5, compress);
		try
		{
			FileUtil.copy(in, out);
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			try
			{
				out.close();
			}
			catch (IOException ex)
			{
				throw new BaseException(ex);
			}
		}
	}

	/**
		Get an <code>OutputStream</code> that can be used
		to upload a physical file for this file item. This method calls
		{@link #getUploadStream(boolean, Boolean)}. A null value is used 
		for the compress parameter to enable auto-compression.
	 	@param checkMd5 TRUE if the Md5 sum should be checked, FALSE otherwise 
	 	@return An output stream to upload file to.
	*/
	public OutputStream getUploadStream(boolean checkMd5)
	{
		return getUploadStream(checkMd5, null);
	}
	
	/**
		Get an <code>OutputStream</code> that can be used
		to upload a physical file for this file item. The client
		application should call <code>close()</code> method to
		finish the upload. Note! Calling this method multiple times
		without closing the returned stream will return the same stream
		and the checkMd5 and compress parameters are ignored.
		
		@param checkMd5 If the MD5 sum of the uploaded data should
			be checked against the MD5 sum already stored in the database.
			If no MD5 exists in the database this parameter is ignored
		@param compress TRUE to store the file in a compressed format, or null
			to use whatever the {@link Directory#getAutoCompress()} is specifying
			or what auto-detection from the {@link #setMimeTypeAuto(String, ItemSubtype)}
			found
		@return An <code>OutputStream</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#WRITE} permission for this item or if the
			file is write protected
		@throws BaseException If there is another error
		@since 2.5
	*/
	public OutputStream getUploadStream(boolean checkMd5, Boolean compress)
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (isWriteProtected() && getLocation().isDownloadable()) 
		{
			throw new PermissionDeniedException("The file is write protected: " + this);
		}
		if(isDetached())
		{
			throw new DetachedItemException("The file cannot be uploaded to a detached item");
		}
		if (uploadStream != null) return uploadStream;
		if (compress == null) compress = uploadCompressed;
		if (newInternalFile == null)
		{
			newInternalFile = getNewFile(compress);
		}
		try
		{
			uploadStream = new UploadStream(newInternalFile, true, checkMd5, compress);
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
		return uploadStream;
	}
	
	/**
		Download the physical file for this file item.
		@param out An <code>OutputStream</code> object which
			will be used to write the file
		@param offset Start reading the physical at the specified offset,
			use 0 to start reading from the beginning
		@throws BaseException If there is an error
	*/
	public void download(OutputStream out, long offset)
		throws BaseException
	{
		InputStream in = getDownloadStream(offset);
		try
		{
			FileUtil.copy(in, out);
			in.close();
			out.flush();
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			try
			{
				in.close();
			}
			catch (IOException e)
			{
				throw new BaseException(e);
			}
		}
	}
	
	/**
		Get an <code>InputStream</code> that can be used
		to download the physical file for this file item. The client
		application should call <code>close()</code> method to
		finish the download. If the file is an external file, the file
		will be loaded from the URL given by {@link #getUrl()}, otherwise
		if the physical file isn't at the primary location an empty stream 
		is returned.
		@param offset Start reading the physical at the specified offset,
			use 0 to start reading from the beginning
		@return An <code>InputStream</code> object
		@throws BaseException If there is an error
	*/
	public InputStream getDownloadStream(long offset)
		throws BaseException
	{
		InputStream download = null;
		Location location = getLocation();
		boolean offsetHandled = false;
		try
		{
			if (location == Location.EXTERNAL)
			{
				ConnectionParameters server = getServerInfo();
				ConnectionManager manager = ConnectionManagerUtil.createConnectionManager(this, server);
				if (manager instanceof ResumableConnectionManager)
				{
					download = ((ResumableConnectionManager)manager).getInputStream(offset);
					offsetHandled = true;
				}
				else
				{
					download = manager.getInputStream();
				}
				setMetadataFromURI(manager.getMetadata());
				if (download == null)
				{
					// For 0-length documents
					download = new ByteArrayInputStream(new byte[0]);
				}
			}
			else if (!location.isDownloadable())
			{
				download = new ByteArrayInputStream(new byte[0]);
				return download;
			}
			else 
			{
				java.io.File file = newInternalFile;
				if (file == null && getInternalName() != null) file = getAbsolutePath(getInternalName());
				if (file == null || !file.exists())
				{
					download = new ByteArrayInputStream(new byte[0]);
					return download;
				}
				else
				{
					download = new BufferedInputStream(new FileInputStream(file));
					if (isCompressed())
					{
						download = new GZIPInputStream(download);
					}
				}
			}
			
			if (offset > 0 && !offsetHandled)
			{
				if (!(download.skip(offset) == offset) && download.read() != -1) 
				{
					throw new IOException("Inputstream-offset error while downloading file: " + getName());
				}
			}
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
		return download;
	}

	/**
		Get a (possible) cached stream for reading from the file. This method
		will enable caching (to the local disk) for external or compressed 
		files. This method is useful if you need to read the file multiple times
		in the same transaction (for example, when trying to auto-detect a file
		format). The cached data is cleared at the end of the transaction.
		
		@return An <code>InputStream</code> object
		@since 2.16
		@see StreamCacher
	*/
	public InputStream getCachedDownloadStream()
	{
		InputStream download = null;
		try
		{
			if (downloadCache == null)
			{
				// Nothing has been cached yet, create a new download
				download = getDownloadStream(0);
				if (getLocation() == Location.EXTERNAL || isCompressed())
				{
					// Only useful to cache external or compressed files
					downloadCache = new StreamCacher(download);
				}
			}
			// If the cache existed or was created use it for the download
			if (downloadCache != null)
			{
				download = downloadCache.getNewStream();
			}
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
		return download;
	}
	
	private static final String separator = java.io.File.separator;
	private static final SimpleDateFormat SUBDIR_FORMAT = 
		new SimpleDateFormat("yyyy-ww"+separator+"dd-ss");

	/**
		Get a new random file path.
		
		@return A new random unique file.
		@throws BaseException If there is an error. 
	*/
	private java.io.File getNewFile(boolean compress)
		throws BaseException
	{
		byte[] random = new byte[4];
		Random generator = new Random();
		String name;
		java.io.File f;
		try
		{
			java.io.File rootDirectory = Application.getUserFilesDirectory();
			String subDirectory = SUBDIR_FORMAT.format(new Date());
			java.io.File directory = new java.io.File(rootDirectory, subDirectory);
			directory.mkdirs();
			if (!directory.exists()) throw new FileNotFoundException(directory.toString());
			do
			{
				StringBuffer sb = new StringBuffer(8);
				generator.nextBytes(random);
				for (int hex : random)
				{
					hex &= 0xFF;
					sb.append(hex < 16 ? "0": "").append(Integer.toHexString(hex));
				}
				name = sb.toString()+(getName() != null ? getName() : "");
				if (name.length() > 30) name = name.substring(0, 30);
				if (compress) name += ".gz";
				f = new java.io.File(directory, name);
			} while (!f.createNewFile());
		}
		catch (IOException ex)
		{
			throw new BaseException("File creation in base.config[userfiles] failed", ex);
		}
		return f;
	}
	
	/**
		Get the relative path of the specified file from the base upload
		directory and convert directory separator to forward slash.
		@since 2.2
		@see #getAbsolutePath(String)
	*/
	static String getRelativePath(java.io.File file)
	{
		String absPath = file.getAbsolutePath();
		String rootPath = Application.getUserFilesDirectory().getAbsolutePath();
		String relPath = absPath.replace(rootPath, "");
		if (relPath.startsWith(separator))
		{
			relPath = relPath.substring(separator.length());
		}
		relPath = relPath.replace(separator, "/");
		return relPath;
	}
	
	/**
		Convert a relative path back to an absolute file path.
		@since 2.2
		@see #getRelativePath(java.io.File)
	*/
	static java.io.File getAbsolutePath(String relativePath)
	{
		// Convert forward slash to file system separator
		relativePath = relativePath.replace("/", separator);
		return new java.io.File(Application.getUserFilesDirectory(), relativePath);
	}
	
	/**
		A fileaction describes if a file should be moved to secondary storage 
		or brought back from it. The actual moving of the file is done by an 
		external progam.  
		
		@author enell
	*/
	public enum Action
	{
		/**
			Do nothing.
		*/
		NOTHING(0),
		/**
			The file should be brought back from secondary storage.
		*/
		MOVE_TO_PRIMARY(1),
		/**
			The file should be moved to secondary storage
		*/
		MOVE_TO_SECONDARY(2);

		/**
			Maps the integer that is stored in the database with an fileaction.
		*/
		private static final Map<Integer, File.Action> valueMapping = new HashMap<Integer, File.Action>();
		
		static
		{
			for(Action action : Action.values())
			{
				Action a = valueMapping.put(action.getValue(), action);
				assert a == null : "Another File.Action with the value "+action.getValue()+" already exists";
			}
		}
		
		/**
			The integer value of this fileaction. 
		*/
		private final int value;
		
		/**
			Creates a new fileaction. Only used internal.
			
			@param value The int value of en action.
		*/
		private Action(int value)
		{
			this.value = value;
		}
		
		/**
			Get the <code>File.Action</code> object when you know the integer code.
			
			@param value The integer value.
			@return The fileaction for the integer value.
		*/
		public static Action fromValue(int value)
		{
			Action action = valueMapping.get(value);
			assert action != null : "action == null for value "+value;
			return action;
		}
		
		/**
			Get the integer value that is used when storing an File.Action to the database.
			@return The intger value of this action. 
		*/
		public int getValue()
		{
			return value;
		}
	}

	/**
		This class is used for uploading a file. It keeps track of the number
		of bytes uploaded and sets the size accordingly.
		@see File#getUploadStream(boolean)
	*/
	private class UploadStream
		extends BufferedOutputStream
	{
		/**
			If this file closed or not.
		*/
		private boolean closed;
		
		/**
			Number of bytes written to the stream.
		*/
		private long totalBytes = 0;
		
		/**
			The md5 sum calculater.
		*/
		private MessageDigest md5 = null;
		
		/**
			The file to write to.
		*/
		private java.io.File file = null;
		
		/**
			If the file should be compressed or not.
		*/
		private boolean compress;
		
		/**
			If the md5 should be checked or not.
		*/
		private boolean checkMd5;

		/**
			Create a new <code>UploadStream</code> and connect it to the
			given file.
			@param file The physical file path to which the stream will write
			@param calculateMd5 TRUE if an md5 sum should be calculated from this uploadstream
			@param checkMd5 If the md5 sum of the file should be checked agains the 
				previously stored md5 sum, useful for making sure secondary storage is
				working properly
			@throws IOException If there is an error during the upload
		*/
		private UploadStream(java.io.File file, boolean calculateMd5, boolean checkMd5, boolean compress)
			throws IOException
		{
			super(compress ? new GZIPOutputStream(new FileOutputStream(file)) : new FileOutputStream(file));
			this.file = file;
			this.compress = compress;
			if (calculateMd5) 
			{
				md5 = MD5.newInstance();
				this.checkMd5 = checkMd5;
			}
			closed = false;
		}
		
		@Override
		public void write(byte[] b, int off, int len)
			throws IOException
		{
			totalBytes += len;
			if (md5 != null) md5.update(b, off, len);
			super.write(b, off, len);
		}
		
		@Override
		public void write(byte[] b)
			throws IOException
		{
			write(b, 0, b.length);
		}

		/**
			Writes the specified byte to this buffered output stream.
			
			@param b the byte to be written.
			@throws IOException if an I/O error occurs.
			@see BufferedOutputStream#write(int)
		*/
		@Override
		public void write(int b)
			throws IOException
		{
			totalBytes++;
			if (md5 != null) md5.update((byte)b);
			super.write(b);
		}

		/**
			Close the stream and register the size and location with the file item.
			
			@throws IOException if an I/O error occurs.
		*/
		@Override
		public void close()
			throws IOException
		{
			if(closed)
			{
				return;
			}
			super.close();
			FileData data = getData();
			data.setSize(totalBytes);
			data.setCompressedSize(file.length());
			data.setCompressed(compress);
			data.setLocation(Location.PRIMARY.getValue());
			data.setRemovedBy(null);
			data.setUrl(null);
			if (md5 != null)
			{
				String oldMd5 = data.getMd5();
				String newMd5 = MD5.getHashString(md5);
				if (checkMd5 && oldMd5 != null && !newMd5.equals(oldMd5))
				{
					throw new InvalidDataException("The new file has a different MD5 checksum (" +
						newMd5+") than the old file ("+oldMd5+").");
				}
				data.setMd5(newMd5);
			}
			uploadStream = null;
			uploadCompressed = false;
			closed = true;
		}
	}
}
