/*
	$Id: PluginDefinition.java 7015 2015-11-17 12:57:57Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.plugin.AnnotationSetterPlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.PluginDefinitionData;
import net.sf.basedb.core.data.PluginPermission;
import net.sf.basedb.core.data.PluginTypeData;
import net.sf.basedb.core.data.GuiContextData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.util.ClassUtil;
import net.sf.basedb.util.JarClassLoader;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.xml.PluginInfo;

import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.EnumSet;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.List;


/**
	A plugin definition represents the executable plugin. Ie. 
	the class name of a class implementing the {@link Plugin}
	interface.

	@author Nicklas, Samuel
	@version 2.0
	@base.modified $Date$
*/
public class PluginDefinition
	extends CommonItem
	implements AnnotatableProxy, Registered
{
	/**
		The type of item represented by this class.
		@see Item#PLUGINDEFINITION Item.PLUGINDEFINITION
		@see #getType() getType()
	*/
	public static final Item TYPE = Item.PLUGINDEFINITION;

	/**
		The maximum length of the class name that can be stored in the
		database. Check the length against this value before calling the
		{@link #setClassName(String) setClassName} method to avoid exceptions.
	*/
	public static final int MAX_CLASSNAME_LENGTH = PluginDefinitionData.MAX_CLASSNAME_LENGTH;
	
	/**
		The maximum length of the jar path name that can be stored in the
		database.
		@since 3.0
	*/
	public static final int MAX_JARFILE_LENGTH = PluginDefinitionData.MAX_JARFILE_LENGTH;

	/**
		The maximum length of the name that can be stored in the
		database.
	*/
	public static final int MAX_NAME_LENGTH = PluginDefinitionData.MAX_NAME_LENGTH;

	/**
		The maximum length of the version that can be stored in the
		database.
	*/
	public static final int MAX_VERSION_LENGTH = PluginDefinitionData.MAX_VERSION_LENGTH;

	/**
		The maximum length of the copyright notice that can be stored in the
		database.
	*/
	public static final int MAX_COPYRIGHT_LENGTH = PluginDefinitionData.MAX_COPYRIGHT_LENGTH;

	/**
		The maximum length of the contact information that can be stored in the
		database.
	*/
	public static final int MAX_CONTACT_LENGTH = PluginDefinitionData.MAX_CONTACT_LENGTH;

	/**
		The maximum length of the email address that can be stored in the
		database.
	*/
	public static final int MAX_EMAIL_LENGTH = PluginDefinitionData.MAX_EMAIL_LENGTH;

	/**
		The maximum length of the url that can be stored in the
		database.
	*/
	public static final int MAX_URL_LENGTH = PluginDefinitionData.MAX_URL_LENGTH;

	/**
		Create a new <code>PluginDefinition</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param className The name of the class that implements the plugin
			interface
		@param jarFile Optional file name of the JAR file which contains the
			plugin class. The JAR file must be located in the plugins.dir 
			directory. If not specified the plugin must be in the classpath 
		@param useRequestedPermissions If the permissions that are returned
			by the {@link Plugin#getPermissions()} method should be used or not.
			If not used the plugin uses the permissions from the logged in user
		@return The new <code>PluginDefinition</code> item
		@throws BaseException If there is an error
	*/
	public static PluginDefinition getNew(DbControl dc, String className, String jarFile, 
		boolean useRequestedPermissions)
		throws BaseException
	{
		PluginDefinition pd = dc.newItem(PluginDefinition.class);
		pd.loadPluginInformation(jarFile, className, useRequestedPermissions);
		pd.getData().setEntryDate(new Date());
		return pd;
	}


	/**
		Get a <code>PluginDefinition</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load.
		@return The <code>PluginDefinition</code> item.
		@throws ItemNotFoundException If an item with the specified ID is not found.
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ READ} permission for the item.
		@throws BaseException If there is another error.
	*/
	public static PluginDefinition getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		PluginDefinition p = dc.loadItem(PluginDefinition.class, id);
		if (p == null) throw new ItemNotFoundException("PluginDefinition[id="+id+"]");
		return p;
	}

	/**
		Get a <code>PluginDefinition</code> item when you know the class name.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param className The class name of the item to load.
		@return The <code>PluginDefinition</code> item.
		@throws ItemNotFoundException If an item with the specified class name is not found.
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ READ} permission for the item.
		@throws BaseException If there is another error.
	*/
	public static PluginDefinition getByClassName(DbControl dc, String className)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (className != null) className = className.trim();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), 
			"GET_PLUGINDEFINITION_FOR_CLASSNAME");
		/*
			SELECT pd
			FROM PluginDefinitionData pd
			WHERE pd.className = :className
		*/
		query.setString("className", className);
		PluginDefinition pd = dc.getItem(PluginDefinition.class, HibernateUtil.loadData(PluginDefinitionData.class, query));
		if (pd == null) throw new ItemNotFoundException("PluginDefinition[className="+className+"]");
		return pd;
	}
	
	
	
	/**
		Get a query that returns plugin definitions.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<PluginDefinition> getQuery()
	{
		return new ItemQuery<PluginDefinition>(PluginDefinition.class);
	}
	
	/**
		Get a query that returns all plugins related
		to a specified GUI context and implementing a certain interface. 
		Ie. The {@link PluginDefinition#isInContext(GuiContext)} returns true for the 
		specified item, and {@link PluginDefinition#supports(PluginType)} returns 
		true for the specified interface.
		
		@param gc The context the plugin should be related to, or null
			if this parameter is irrelevant
		@param interfaceName The complete name of the interface that the plugin
			must implement, or null if this parameter is irrelevant
	*/
	public static ItemQuery<PluginDefinition> getQuery(GuiContext gc, String interfaceName)
	{
		ItemQuery<PluginDefinition> query = getQuery();
		Item item = gc == null ? null : gc.getItem();
		GuiContext.Type type = gc == null ? null : gc.getType();
		if (item != null || type != null)
		{
			query.joinPermanent(Hql.innerJoin("guiContexts", "gcx"));
			if (item != null)
			{
				query.restrictPermanent(
					Restrictions.eq(
						Hql.property("gcx", "itemType"), 
						Expressions.integer(item.getValue())
					)
				);
			}
			if (type != null)
			{
				query.restrictPermanent(
					Restrictions.eq(
						Hql.property("gcx", "contextType"), 
						Expressions.integer(type.getValue())
					)
				);
			}
		}
		if (interfaceName != null)
		{
			interfaceName = interfaceName.trim();
			query.joinPermanent(Hql.innerJoin("pluginTypes", Item.PLUGINTYPE.getAlias()));
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property(Item.PLUGINTYPE.getAlias(), "interfaceName"), 
					Expressions.parameter("interfaceName")
				)
			);
			query.setPermanentParameter("interfaceName", interfaceName, Type.STRING);
		}
		return query;
	}
	
	/**
		Count the number of plugins that can be used in a given context.
		@param dc An open DbControl object
		@param gc The context to count plugins for
		@return A map with the main type as the key and the number of
			plugins as the value
	 	@throws BaseException If the plugins could not be counted. 
	*/
	public static Map<Plugin.MainType, Integer> countPlugins(DbControl dc, GuiContext gc)
		throws BaseException
	{
		SessionControl sc = dc.getSessionControl();
		boolean hasConfigPermission = sc.hasPermission(Permission.USE, Item.PLUGINCONFIGURATION);
		boolean hasPluginPermission = sc.hasPermission(Permission.USE, Item.PLUGINDEFINITION);
		
		org.hibernate.Session session = dc.getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "COUNT_PLUGINS_BY_TYPE_FOR_CONTEXT");
		/*
			SELECT plg.mainType, count(DISTINCT plg.id)
			FROM PluginDefinitionData plg
			JOIN plg.guiContexts gcx
			LEFT JOIN plg.configurations cfg
			WITH true = :hasConfigPermission OR cfg.owner.id = :owner 
				OR cfg.itemKey.id IN (:itemKeys) OR cfg.projectKey.id IN (:projectKeys)
			WHERE 
				(true = :hasPluginPermission OR plg.owner.id = :owner 
				OR plg.itemKey.id IN (:itemKeys) OR plg.projectKey.id IN (:projectKeys))
			AND gcx.itemType = :itemType AND gcx.contextType = :contextType
			AND (
				(plg.requiresConfiguration = false)
				OR
				(plg.requiresConfiguration = true AND NOT cfg.id IS NULL)
			)
			AND plg.disabled = false
			GROUP BY plg.mainType
		*/
		
		// Context
		query.setInteger("itemType", gc.getItem().getValue());
		query.setInteger("contextType", gc.getType().getValue());

		// Role permissions
		query.setBoolean("hasConfigPermission", hasConfigPermission);
		query.setBoolean("hasPluginPermission", hasPluginPermission);
		
		// Owner
		query.setInteger("owner", sc.getLoggedInUserId());

		// Shared keys
		Set<Integer> ZERO_SET = Collections.singleton(0);
		Set<Integer> itemKeys = sc.getItemKeys(Permission.USE);
		if (itemKeys == null || itemKeys.size() == 0) itemKeys = ZERO_SET;
		query.setParameterList("itemKeys", itemKeys, TypeWrapper.INTEGER.getHibernateType());

		// Project keys for the active project
		Set<Integer> projectKeys = null;
		if (sc.getActiveProjectId() != 0) projectKeys = sc.getProjectKeys(Permission.USE);
		if (projectKeys == null || projectKeys.size() == 0) projectKeys = ZERO_SET;
		query.setParameterList("projectKeys", projectKeys, TypeWrapper.INTEGER.getHibernateType());
	
		List<Object> pluginCount = HibernateUtil.loadList(Object.class, query, dc.getSessionControl());
		Map<Plugin.MainType, Integer> result = new HashMap<Plugin.MainType, Integer>(5);
		for (Object cnt : pluginCount)
		{
			Object[] o = (Object[])cnt;
			result.put(Plugin.MainType.fromValue((Integer)o[0]), ((Long)o[1]).intValue());
		}
		
		return result;
	}
	
	/**
		Check the installation status of the given list of
		plug-ins.
		@param dc An optional DbControl (if null a new internal connection will be used)
		@param plugins An array with plug-in information objects
		@since 3.0
	*/
	public static void checkInstallation(DbControl dc, Collection<PluginInfo> plugins)
	{
		if (plugins == null || plugins.size() == 0) return;
		
		org.hibernate.Session session = dc == null ? HibernateUtil.newSession() : dc.getHibernateSession();
		try
		{
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_PLUGINDEFINITION_FOR_CLASSNAME");
			
			for (PluginInfo info : plugins)
			{
				query.setString("className", info.getClassName());
				PluginDefinitionData plugin = HibernateUtil.loadData(PluginDefinitionData.class, query);
				if (plugin != null)
				{
					info.setInternalId(plugin.getId());
					info.setDisabled(plugin.isDisabled());
					info.setType(Plugin.MainType.fromValue(plugin.getMainType()));
				}
				else
				{
					info.setNotFound(true);
				}
			}
		}
		finally
		{
			if (dc == null)
			{
				HibernateUtil.close(session);
			}
		}	
	}
	
	/**
		Install or update the plug-in definition with information from the info
		object. Existing plug-ins will be updated and the disabled flag is cleared
		(unless it is deprecated).
		New plug-ins are configured with additional options from the info object:
		<ul>
		<li>max-memory: Maximum memory in bytes when the plug-in is executed as an external
			process by a job agent
		<li>immediate-execution: If the plug-in is allowed to bypass the job queue.
		<li>everyone-use: If the plug-in should be shared with USE permission to everyone.
		<li>deprecated: Disable the plug-in if it is already installed, do not install
			if it is missing
		</ul>
		@return The plugin definition or null if no plug-in was installed
		@since 3.0
	*/
	public static PluginDefinition installOrUpdate(DbControl dc, PluginInfo info, String jarFile, ItemKey shareToEveryone)
	{
		PluginDefinition plugin = null;
		String className = info.getClassName();
		boolean deprecated = Values.getBoolean(info.getProperty("deprecated"));
		try
		{
			plugin = PluginDefinition.getByClassName(dc, className);
			plugin.loadPluginInformation(jarFile, className, true);
		}
		catch (ItemNotFoundException ex)
		{
			if (!deprecated)
			{
				plugin = PluginDefinition.getNew(dc, className, jarFile, true);
				plugin.setMaxMemory(Values.getLong(info.getProperty("max-memory"), null));
				plugin.setAllowImmediateExecution(Values.getBoolean(info.getProperty("immediate-execution")));
				if (Values.getBoolean(info.getProperty("everyone-use")))
				{
					plugin.setItemKey(shareToEveryone);
				}
				dc.saveItem(plugin);
				plugin.info = info; // So that we can set an ID on the info object after commit.
			}
		}
		if (plugin != null)
		{
			plugin.setAbout(info.getAbout(), true);
			plugin.setDisabled(deprecated);
			info.setType(plugin.getMainType());
		}
		return plugin;
	}
	
	private PluginInfo info;
	
	PluginDefinition(PluginDefinitionData pd)
	{
		super(pd);
	}
	
	@Override
	PluginDefinitionData getData()
	{
		return (PluginDefinitionData)getBasicData();
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	@Override
	public boolean isRemoved()
	{
		return getData().getRemovedBy() != null;
	}
	@Override
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemovedBy(removed ? getSessionControl().getLoggedInUserId() : null);
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if a {@link PluginConfiguration} is using this definition.
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLUGINCONFIGURATIONS_FOR_PLUGINDEFINITION", "count(*)");
		/*
			SELECT {1}
			FROM PluginConfigurationData pc
			WHERE pc.pluginDefinition = :pluginDefinition
		*/
		query.setEntity("pluginDefinition", this.getData());
		long count = HibernateUtil.loadData(Long.class, query);
		if (count == 0)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_JOBS_FOR_PLUGINDEFINITION", "count(*)");
			/*
				SELECT {1}
				FROM JobData j
				WHERE j.pluginDefinition = :pluginDefinition
			*/
			query.setEntity("pluginDefinition", this.getData());
			count = HibernateUtil.loadData(Long.class, query);
		}
		return count > 0 || super.isUsed();
	}
	
	@Override
	void onAfterInsert() 
		throws BaseException 
	{
		super.onAfterInsert();
		if (info != null) 
		{
			info.setInternalId(getId());
		}
	}


	/**
	 Get all:
		<ul>
		<li>{@link PluginConfiguration}:s created from this plugin
		<li>{@link Job}:s using this plugin
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();

		// Plugin configurations
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLUGINCONFIGURATIONS_FOR_PLUGINDEFINITION", "pc.id");
			/*
				SELECT {1}
				FROM PluginConfigurationData pc
				WHERE pc.pluginDefinition = :pluginDefinition
			*/
		query.setEntity("pluginDefinition", this.getData());
		addUsingItems(using, Item.PLUGINCONFIGURATION, query);

		// Jobs
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_JOBS_FOR_PLUGINDEFINITION", "j.id");
			/*
				SELECT {1}
				FROM JobData j
				WHERE j.pluginDefinition = :pluginDefinition
			*/
		query.setEntity("pluginDefinition", this.getData());
		addUsingItems(using, Item.JOB, query);
		
		return using;
	}
	// -------------------------------------------

	/*
		From the Annotatable & AnnotatableProxy interface
		-------------------------------------------
	*/
	/**
		@since 2.5
	*/
	@Override
	public AnnotationSet getAnnotationSet()
		throws PermissionDeniedException, BaseException
	{
		DbControl dc = getDbControl();
		AnnotationSet as = dc.getItem(AnnotationSet.class, getData().getAnnotationSet(), this);
		if (as == null)
		{
			as = AnnotationSet.getNew(dc, this);
			getData().setAnnotationSet(as.getData());
			dc.saveItemIf(this, as, false);
		}
		return as;
	}
	/**
		@since 2.5
	*/
	@Override
	public boolean isAnnotated()
	{
		return getData().getAnnotationSet() != null;
	}
	/**
		@since 2.5
	*/
	@Override
	public void removeAnnotations()
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		AnnotationSetData as = getData().getAnnotationSet();
		if (as != null)
		{
			HibernateUtil.deleteData(getDbControl().getHibernateSession(), as);
			getData().setAnnotationSet(null);
		}
	}
	/**
		@return Always null
		@since 2.5
	*/
	@Override
	public Protocol getProtocol()
	{
		return null;
	}
	/**
		@return Always null
		@since 2.5
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
	{
		return null;
	}
	/**
		Get the annotation types that can be used by this configuration. This method
		requires that the plug-in implements the {@link AnnotationSetterPlugin} interface.
		@return A query or null if this plug-in doesn't use annotations
		@since 2.5
	*/
	@Override
	public ItemQuery<AnnotationType> getAnnotationTypes()
	{
		Plugin p = newInstance(Plugin.class, getSessionControl(), null, null);
		ItemQuery<AnnotationType> query = null;
		if (p instanceof AnnotationSetterPlugin)
		{
			query = ((AnnotationSetterPlugin)p).getAnnotationTypes();
		}
		return query;
	}
	/**
		Get a message explaining what the proxy does with the annotations.
		@return A message or null
		@since 2.5
	*/
	@Override
	public String getAnnotationMessage()
	{
		Plugin p = newInstance(Plugin.class, getSessionControl(), null, null);
		String message = null;
		if (p instanceof AnnotationSetterPlugin)
		{
			message = ((AnnotationSetterPlugin)p).getAnnotationMessage();
		}
		return message;
	}	
	// -------------------------------------------	
	
	/**
		Get the class name of the Java class that implements this plugin. The
		specified class must implement the {@link net.sf.basedb.core.plugin.Plugin}
		interface.
		@return The class name for this plugin
	*/
	public String getClassName()
	{
		return getData().getClassName();
	}
	/**
		The class name cannot be changed. Used internally only.
	*/
	private void setClassName(String className)
		throws InvalidDataException
	{
		getData().setClassName(StringUtil.setNotNullString(className, "className", MAX_CLASSNAME_LENGTH));
	}

	/**
		Get the path to the JAR file that contains the plugin class file.
		If this value is null, the plugin must be on the classpath.
	*/
	public String getJarPath()
	{
		String file = getJarFile();
		if (file == null) return null;
		return new java.io.File(Application.getPluginsDirectory(), file).getAbsolutePath();
	}
	/**
		The jar path cannot be changed. Used internally only.
	*/
	private void setJarFile(String jarFile)
		throws InvalidDataException
	{
		getData().setJarFile(StringUtil.setNullableString(jarFile, "jarFile", MAX_JARFILE_LENGTH));
	}
	
	/**
		Get the name of the JAR file that contains this plug-in.
		If this value is null, the plug-in must be on the classpath
		(eg. in WEB-INF/lib).
		@since 3.0
	*/
	public String getJarFile()
	{
		return getData().getJarFile();
	}
	
	/**
		Get the version of the plugin.
		@return A string with the version of the plugin, or null if not known
	*/
	public String getVersionString()
	{
		return getData().getVersionString();
	}
	/**
		Update version information for this plug-in.
		@throws InvalidDataException If the version is too long
		@since 3.0
	*/
	private void setVersionString(String version)
		throws InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setVersionString(StringUtil.setNullableString(version, "version", MAX_VERSION_LENGTH));
	}
	
	/**
		Get a copyright notice for the plugin.
		@return A string with a copyright notice, or null if not known
	*/
	public String getCopyright()
	{
		return getData().getCopyright();
	}
	/**
		Update copyright information about this plug-in.
		@throws InvalidDataException If the copyright is too long
		@since 3.0
	*/
	private void setCopyright(String copyright)
		throws InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setCopyright(StringUtil.setNullableString(copyright, "copyright", MAX_COPYRIGHT_LENGTH));
	}

	/**
		Get contact information for the plugin.
		@return A string with contact information, or null if not known
	*/
	public String getContact()
	{
		return getData().getContact();
	}
	/**
		Set contact information for this plug-in.
		@throws InvalidDataException If the contact is too long
		@since 3.0
	*/
	private void setContact(String contact)
		throws InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setContact(StringUtil.setNullableString(contact, "contact", MAX_CONTACT_LENGTH));
	}
	
	/**
		Get an email address that can be used to get more information about the plugin.
		@return A string with the email address, or null if not known
	*/
	public String getEmail()
	{
		return getData().getEmail();
	}
	/**
		Set the email address that can be used to contact the
		developer of the plug-in.

		@throws InvalidDataException If the email is too long
		@since 3.0
	*/
	private void setEmail(String email)
		throws InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setEmail(StringUtil.setNullableString(email, "email", MAX_EMAIL_LENGTH));
	}

	/**
		Get a URL with more information about the plugin.
		@return A string containing the URL, or null if not known
	*/
	public String getUrl()
	{
		return getData().getUrl();
	}
	/**
		Set the URL to a web page with more information about the plug-in.
		@throws InvalidDataException If the url is too long
		@since 3.0
	*/
	public void setUrl(String url)
		throws InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setUrl(StringUtil.setNullableString(url, "url", MAX_URL_LENGTH));
	}

	/**
		Update the plug-in information with the information in the
		about object.
		@param about An About object
		@param force If TRUE all values are updated, if FALSE only non-null
			values
		@since 3.0
	*/
	public void setAbout(About about, boolean force)
	{
		if (force || about.getName() != null) setName(about.getName());
		if (force || about.getDescription() != null) setDescription(about.getDescription());
		if (force || about.getVersion() != null) setVersionString(about.getVersion());
		if (force || about.getCopyright() != null) setCopyright(about.getCopyright());
		if (force || about.getContact() != null) setContact(about.getContact());
		if (force || about.getEmail() != null) setEmail(about.getEmail());
		if (force || about.getUrl() != null) setUrl(about.getUrl());
	}
	
	/**
		Checks if this plugin is disabled or not. A disabled plugin
		can't be used.
		@since 3.0
	*/
	public boolean isDisabled()
	{
		return getData().isDisabled();
	}
	
	/**
		Disabled or enable this plugin.
		@since 3.0
	*/
	public void setDisabled(boolean disabled)
	{
		checkPermission(Permission.WRITE);
		getData().setDisabled(disabled);
	}
	
	/**
		If this plugin is trusted or not. A trusted plugin is executed without any
		restrictions, an untrusted have restrictions similar to an applet in a 
		web browser. NOTE!!! CURRENTLY NOT IMPLEMENTED.
		@return TRUE if the plugin is trusted, FALSE otherwise
	*/
	public boolean isTrusted()
	{
		return getData().isTrusted();
	}
	
	/**
		Set if this plugin should be trusted or not. NOTE!!! CURRENTLY NOT IMPLEMENTED.
		@see #isTrusted()
		@throws PermissionDeniedException If the logged in user doesn't have write
			permission
	*/
	public void setTrusted(boolean trusted)
	{
		checkPermission(Permission.WRITE);
		getData().setTrusted(trusted);
	}
	
	/**
		Get the maximum amount of memory the plugins is allowed to use. The returned value
		is only a guideline to the job agent which may choose to use a lower setting due
		to the servers limitations. A null value means the job agent is free to choose
		any amount. Usually, it will select the default value as specified by the
		<code>agent.maxMemory</code> setting in it's configuration file.
		@return The maximum amount of memory to use, or null
	*/
	public Long getMaxMemory()
	{
		return getData().getMaxMemory();
	}
	
	/**
		Set the maximum amount of memory the plugin is allowed to use. 
		@see #getMaxMemory()
		@param maxMemory The maximum amount, or null if the job agent should decide
		@throws PermissionDeniedException If the logged in user doesn't have write
			permission
	*/
	public void setMaxMemory(Long maxMemory)
	{
		checkPermission(Permission.WRITE);
		getData().setMaxMemory(maxMemory);
	}
	
	/**
		Get the type of plugin.
		@return A type code that is defined in the enum 
			{@link net.sf.basedb.core.plugin.Plugin.MainType}.
	*/
	public Plugin.MainType getMainType()
	{
		return Plugin.MainType.fromValue(getData().getMainType());
	}
	/**
		The type cannot be changed. Used internally only.

		@base.developer
		This value is initialised upon installation to: plugin.getMainType()
		See: {@link #loadPluginInformation(String, String, boolean)}
	
		@throws InvalidDataException If the type is null
	*/
	private void setMainType(Plugin.MainType type)
		throws InvalidDataException
	{
		if (type == null) throw new InvalidUseOfNullException("type");
		getData().setMainType(type.getValue());
	}
	
	/**
		Check if the plugin is interactive or not. An interactive
		plugin implements the {@link InteractivePlugin} interface
		and can be dynamically configured by telling the client
		application what configuration parameters it needs.
	*/
	public boolean isInteractive()
	{
		return getData().isInteractive();
	}
	
	private void setInteractive(boolean interactive)
	{
		getData().setInteractive(interactive);
	}

	/**
		Check if the plugin supports configurations or not when executing a job. 
		Note! A false return value doesn't mean that the core stops anyone from creating a
		{@link PluginConfiguration} for this plugin. The return value
		should be used by client applications for better GUI.
	*/
	public boolean supportsConfigurations()
	{
		return getData().getSupportsConfigurations();
	}
	private void setSupportsConfigurations(boolean supportsConfigurations)
	{
		getData().setSupportsConfigurations(supportsConfigurations);
	}

	/**
		Check if the plugin requires a configuration to execute a job.
		Note! The core will check the return value of this method when 
		creating a job and throw an {@link InvalidDataException} if
		a configuration is required but isn't supplied.
	*/
	public boolean requiresConfiguration()
	{
		return getData().getRequiresConfiguration();
	}
	private void setRequiresConfiguration(boolean requiresConfiguration)
	{
		getData().setRequiresConfiguration(requiresConfiguration);
	}
	
	/**
	  	Check if this plugin should execute with or without permissions.
	  	Note! The plugin will use the permissions of the logged in user if 
	  	this is set to FALSE which could lead to this plugin not will be 
	  	able to save/create files if the user doesn't have that permission.
	  	To be sure that the plugin will execute under certain permissions
	  	this should be set to TRUE.
	  	@see #setPermissions(RoleKey, Set, Set)
	*/
	public boolean getUsePermissions()
	{
		return getData().getUsePermissions();
	}
	public void setUsePermissions(boolean usePermissions)
	{
		checkPermission(Permission.SET_PERMISSION);
		getData().setUsePermissions(usePermissions);
	}
	
	/**
		If immediate execution of jobs are allowed or if they always must go through
		the job queue. Immediate execution is needed by export plugins that want
		to use the immediate download feature.
		@return TRUE if immediate execition is allowed, FALSE it jobs must be
			added to the job queue
		@since 2.2
		@see net.sf.basedb.core.plugin.ImmediateDownloadExporter
	*/
	public boolean getAllowImmediateExecution()
	{
		return getData().getAllowImmediateExecution();
	}
	public void setAllowImmediateExecution(boolean allow)
	{
		checkPermission(Permission.WRITE);
		getData().setAllowImmediateExecution(allow);
	}
	
	/**
		If this plug-in can be executed by the internal job queue or
		not. If not, the plug-in must be configured to be executed by a 
		job agent. NOTE! The plug-in can always be excuted by job agents,
		they ignore this value.
		
		@return TRUE if the internal job queue is used, FALSE if not
		@since 2.5
	*/
	public boolean getUseInternalJobQueue()
	{
		return getData().getUseInternalJobQueue();
	}
	/**
		Set if the plug-in can use the internal job queue or not.
		@param use TRUE to use the internal job queue, FALSE to only use
			job agents
		@since 2.5
	*/
	public void setUseInternalJobQueue(boolean use)
	{
		checkPermission(Permission.WRITE);
		getData().setUseInternalJobQueue(use);
	}
	
	/**
		Check if it makes sense to use this plugin in the context
		of the specified item. For example, a plugin that imports
		reporters would return true if passed {@link Item#REPORTER}.
		@param context The {@link GuiContext} type
		@return TRUE if the plugin handles the item, FALSE otherwise
	*/
	public boolean isInContext(GuiContext context)
	{
		if (context == null) return false;
		GuiContextData gc = new GuiContextData(context.getItem().getValue(), context.getType().getValue());
		return getData().getGuiContexts().contains(gc);
	}
	
	/**
		Get the {@link Item} types where it makes sense to use this plugin
		for a client application. For example, a plugin that imports
		reporters would return a set containing {@link Item#REPORTER}.
		@return A <code>Set</code> of <code>Item</code>:s or an empty set
			if the plugin is not concerned about items
	*/
	public Set<GuiContext> getGuiContexts()
	{
		Set<GuiContextData> guiContexts = getData().getGuiContexts();
		Set<GuiContext> pluginContexts = new HashSet<GuiContext>(guiContexts.size());
		for (GuiContextData gc : guiContexts)
		{
			pluginContexts.add(
				new GuiContext(Item.fromValue(gc.getItemType()), GuiContext.Type.fromValue(gc.getContextType()))
			);
		}
		return pluginContexts;
	}
	
	/**
		Set the contexts where this plugin can be used.
	*/
	private void setGuiContexts(Set<GuiContext> pluginContexts)
	{
		if (pluginContexts == null) return;
		Set<GuiContextData> guiContexts = getData().getGuiContexts();
		guiContexts.clear();
		for (GuiContext gc : pluginContexts)
		{
			guiContexts.add(new GuiContextData(gc.getItem().getValue(), gc.getType().getValue()));
		}
	}
	
	/**
		Create a new {@link PluginConfiguration} for this plugin.
		@return The new <code>PluginConfiguration</code> item
		@throws PermissionDeniedException If
			the logged in user doesn't have use
			permission for this plugin definition
		@throws BaseException If there is another
			error
	*/
	public PluginConfiguration newPluginConfiguration()
		throws PermissionDeniedException, BaseException
	{
		PluginConfiguration pc = PluginConfiguration.getNew(getDbControl(), this);
		return pc;
	}

	/**
		Get a query that returns the configurations 
		for this plugin definition.

		@return An {@link ItemQuery} object
	*/
	public ItemQuery<PluginConfiguration> getPluginConfigurations()
	{
		ItemQuery<PluginConfiguration> query = PluginConfiguration.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("pluginDefinition"), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Get a query that returns the plugin types 
		for this plugin definition.

		@return An {@link ItemQuery} object
	*/
	public ItemQuery<PluginType> getPluginTypes()
	{
		ItemQuery<PluginType> query = PluginType.getQuery();
		query.joinPermanent(Hql.innerJoin("plugins", Item.PLUGINDEFINITION.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.PLUGINDEFINITION.getAlias()), 
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Check if a plugin implements the interface specified by the plugin
		type.
		@param pluginType The pluginType to check
		@return TRUE if the plugin implements the specified interface, FALSE otherwise
	*/
	public boolean supports(PluginType pluginType)
	{
		return getData().getPluginTypes().contains(pluginType.getData());
	}
	
	/**
		Check if a plugin implements the interface specified by the plugin
		type.
		@param interfaceName The pluginType to check
		@return TRUE if the plugin implements the specified interface, FALSE otherwise
	*/
	public boolean supports(String interfaceName)
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(
			getDbControl().getHibernateSession(), "GET_PLUGINTYPE_FOR_INTERFACENAME");
		query.setString("interfaceName", interfaceName);
		PluginTypeData pt = HibernateUtil.loadData(PluginTypeData.class, query);
		return getData().getPluginTypes().contains(pt);
	}

	
	/**
		Get an unitialized instance of the plugin.
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#USE} permission to the items.
		@throws BaseException If there is a problem creating an instance.
	*/
	public Plugin newInstance()
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.USE);
		if (isDisabled())
		{
			throw new PermissionDeniedException("The plugin is disabled: " + getName());
		}
		return newInstance(getJarPath(), getClassName(), null, false);
	}
	
	/**
		Get an unitialized instance of the plugin implementation.
	 	@param <P> The kind of returned object that extends Plugin. 
		@param clazz The class of the plugin which must be a subclass of <code>Plugin</code>
		@return An uninitialised plugin instance
	 */
	public <P extends Plugin> P newInstance(Class<P> clazz)
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.USE);
		if (isDisabled())
		{
			throw new PermissionDeniedException("The plugin is disabled: " + getName());
		}
		ClassLoader alternateClassLoader = null;
		if (clazz.getClassLoader() != this.getClass().getClassLoader())
		{
			alternateClassLoader = clazz.getClassLoader();
		}
		Plugin p = newInstance(getJarPath(), getClassName(), alternateClassLoader, false);
		if (!clazz.isInstance(p))
		{
			throw new ClassCastException("Class '" + p.getClass().getName() + "' doesn't implement '" + clazz.getName() + "'");
		}
		return clazz.cast(p);
	}

	
	/**
		Get an initialized instance of the plugin implementation.
	 	@param <P> The kind of returned object that extends Plugin. 
		@param clazz The class of the plugin which must be a subclass of <code>Plugin</code>
		@param sc The session control to initialise the plugin with
		@param configuration The configuration parameters to initialise the plugin with,
			or null if no configuration exists
		@param job The job parameters to intialise the plugin with, or null if no
			job exists
		@return An initialised plugin instance
	*/
	public <P extends Plugin> P newInstance(Class<P> clazz, SessionControl sc, PluginConfiguration configuration, Job job)
		throws PermissionDeniedException, BaseException
	{
		P p = newInstance(clazz);
		if (sc == null) sc = getSessionControl();
		p.init(sc.getPluginSessionControl(this, job), 
			configuration == null ? new ParameterValuesImpl(this) : configuration.getParameterValuesImpl(configuration.getParameterVersion()),
			job == null ? new ParameterValuesImpl(this) : job.getParameterValuesImpl()
		);
		return clazz.cast(p);
	}

	/**
		Create a new object instance of the given class.

		@param jarPath The full path to the JAR file
		@param className The class to create
		@param unloadBefore If the class definition should be unloaded before the new
			instance is created, unloading may not work if a client
			program holds a reference to an instance of the class
		@return The created instance
	*/
	private Plugin newInstance(String jarPath, String className, ClassLoader alternateClassLoader, boolean unloadBefore)
		throws BaseException
	{
		try
		{
			Class<?> c = null;
			if (alternateClassLoader != null)
			{
				try
				{
					c = ClassUtil.checkAndLoadClass(alternateClassLoader, className, true, Plugin.class);
				}
				catch (Exception ex)
				{}
			}
			if (jarPath == null)
			{
				c = Class.forName(className);
			}
			else
			{
				ClassLoader cl = JarClassLoader.getInstance(jarPath, unloadBefore || Application.autoUnloadPlugins());
				c = ClassUtil.checkAndLoadClass(cl, className, true, Plugin.class);
			}
			return (Plugin)c.newInstance();
		}
		catch (BaseException bex)
		{
			throw bex;
		}
		catch (Exception ex)
		{
			throw new BaseException("Could not instantiate plugin class: "+className+"", ex);
		}
	}

	/**
	  	Sets the permissions for this plugin. It is not possible to give the plugin more
	  	permissions than the logged in user have. Permissions that are neither granted nor
	  	denied are checked against the permissions for the user that is running the plugin.
	  	<p>
	  	NOTE! The granted permissions are always granted, regardless of the logged in
	  	user's permissions. The denied permissions are always denied regardless of the
	  	logged in user's permissions. This applies even to the root user account.
	  	<p>
	  	If the granted permissions is null or empty and the denied permissions
	  	is null or contains the same set of permissions as defined by the item type
	  	the permissions are removed from the database, which means that only those
	  	permissions which differs from the defaults are stored in the database.
	  	
	  	@param roleKey The rolekey for this plugin
	  	@param granted The permission to grant to the plugin
	  	@param denied The permissions to deny the plugin
	  	@throws PermissionDeniedException If the logged in user doesn't have {@link Permission#SET_PERMISSION} 
	  		for this <code>PluginDefinition</code> or trying to grant the plugin more permission then the logged in user
	  	@throws InvalidUseOfNullException If the rolekey is null
	*/
	public void setPermissions(RoleKey roleKey, Set<Permission> granted,  Set<Permission> denied)
		throws PermissionDeniedException, InvalidUseOfNullException
	{
		checkPermission(Permission.SET_PERMISSION);
		if (roleKey == null) throw new InvalidUseOfNullException("roleKey");
		Item itemType = roleKey.getItemType();
		if (itemType.getDefinedPermissions() == null)
		{
			throw new PermissionDeniedException(
					"Not allowed to set permissions for item type '" + 
					itemType.name() + "' since it has no key.");
		}
		
		if ((granted == null || granted.size() == 0) && 
			(denied == null || denied.containsAll(itemType.getDefinedPermissions())))
		{
			getData().getPermissions().remove(roleKey.getData());
		}
		else
		{
			if (granted != null)
			{
				for (Permission p : granted)
				{
					if (!getSessionControl().hasPermission(p, itemType))
					{
						throw new PermissionDeniedException("Not allowed to set " + p + 
							" permissions for '" + itemType.name() + "' to this plugin: " + this);
					}
				}
			}
			int iGranted = Permission.grant(granted);
			int iDenied = Permission.deny(denied) & ~iGranted;
			
			getData().getPermissions().put(roleKey.getData(), 
				new PluginPermission(iGranted, iDenied));
		}
	}
	
	/**
	 	Get the permissions that have been granted to this plugin.
	 	
	  	@param roleKey The role key to get the permission for
	  	@return A set containing the granted permissions for this plugin, the
	  		set is empty if no permissions has been granted
	  	@throws InvalidUseOfNullException If the role key is null
	*/
	public Set<Permission> getGranted(RoleKey roleKey)
		throws InvalidUseOfNullException
	{
		if (roleKey == null) throw new InvalidUseOfNullException("roleKey");
		PluginPermission pp = getData().getPermissions().get(roleKey.getData());
		return pp != null ? Permission.fromInt(pp.getGranted()) : EnumSet.noneOf(Permission.class);
	}
	
	/**
	 	Get the permissions that have been denied to this plugin.
	  	@param roleKey The role key to get the permission for
	  	@return A set containing the denied permissions for this plugin, the
	  		set is empty if no permissions has been denied
	  	@throws InvalidUseOfNullException If the role key is null
	*/
	public Set<Permission> getDenied(RoleKey roleKey)
		throws InvalidUseOfNullException
	{
		if (roleKey == null) throw new InvalidUseOfNullException("roleKey");
		PluginPermission pp = getData().getPermissions().get(roleKey.getData());
		return pp != null ? Permission.fromDeniedInt(pp.getDenied()) : EnumSet.allOf(Permission.class);
	}
	
	/**
		Load a plugin and copy information to the internal data structure.
	 	@param jarFile File name of the jar file containing the plug-in. The
	 		JAR file must be located in the directory spercified by 'plugins.dir' setting
	 		in base.config. Null is allowed if plugin is located in the classpath (eg. WEB-INF/lib).
	 	@param className Full class name of the plugin's main class.
	 	@param useRequstedPermissions TRUE if plugin should use the permissions it requests. 
	 	@throws PermissionDeniedException If logged in user doesn't have write permission to this item.
	 	@throws InvalidDataException If any of the arguments is malformed or 
	 		if setting the plugin's properties fails.
	 	@throws BaseException If something else fails.
	*/
	public void loadPluginInformation(String jarFile, String className, boolean useRequstedPermissions)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (jarFile != null)
		{
			if (!Path.isValidName(jarFile))
			{
				throw new InvalidDataException("JAR file name is not a valid name: " + jarFile);
			}
		}
		
		String jarPath = jarFile == null ? 
				null : new java.io.File(Application.getPluginsDirectory(), jarFile).getAbsolutePath();
		Plugin plugin = newInstance(jarPath, className, null, true);

		// The plugin type
		setMainType(plugin.getMainType());

		// Configurations support
		setSupportsConfigurations(plugin.supportsConfigurations());
		setRequiresConfiguration(supportsConfigurations() && plugin.requiresConfiguration());
		
		// The name, class name and jar filename
		if (getName() == null) setName(plugin.getClass().getSimpleName());
		setClassName(className);
		setJarFile(jarFile);
		
		if (plugin instanceof InteractivePlugin)
		{
			InteractivePlugin iPlugin = (InteractivePlugin)plugin;
			setGuiContexts(iPlugin.getGuiContexts());
			setInteractive(true);
		}
		else
		{
			setInteractive(false);
		}
		
		// Find which plugin type interfaces the plugin implements
		Set<Class<?>> interfaces = ClassUtil.getInterfaces(plugin.getClass());
		Set<String> interfaceNames = new HashSet<String>(interfaces.size());
		for (Class<?> in : interfaces)
		{
			interfaceNames.add(in.getName());
		}
		org.hibernate.Query query = HibernateUtil.createQuery(
			getDbControl().getHibernateSession(), "SELECT pt FROM PluginTypeData pt");
		List<PluginTypeData> types = HibernateUtil.loadList(PluginTypeData.class, query, null);

		for (PluginTypeData pt : types)
		{
			if (interfaceNames.contains(pt.getInterfaceName()))
			{
				getData().getPluginTypes().add(pt);
			}
		}
		
		// Update plugin permissions
		if (useRequstedPermissions)
		{
			Collection<Permissions> permissions = plugin.getPermissions();
			if (permissions == null || permissions.size() == 0)
			{
				setUsePermissions(false);
			}
			else
			{
				setUsePermissions(true);
				DbControl dc = getDbControl();
				for (Permissions p : permissions)
				{
					Item itemType = p.getItemType();
					boolean allowed = itemType.getDefinedPermissions() != null;
					if (allowed)
					{
						try
						{
							RoleKey roleKey = RoleKey.getByItemType(dc, itemType);
							// Translate the "maybe granted" permissions to denied permissions
							Set<Permission> denied = Permission.fromDeniedInt(~Permission.grant(p.getMaybeGranted()));
							setPermissions(roleKey, p.getAlwaysGranted(), denied);
						}
						catch (ItemNotFoundException ex)
						{
							allowed = false;
						}
					}
					if (!allowed)
					{
						throw new PermissionDeniedException(
								"Not allowed to set permissions for item type '" + 
								itemType.name() + "' since it has no key.");
					}
					
				}
			}
		}
	}	
}
