/*
	$Id $

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.authentication;

import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.core.SessionControl;

/**
	Holds information about the method that was used to verify a user when logging in.
	See {@link SessionControl#getAuthenticationMethod()}. The {@link #INTERNAL}
	is method is a static constant used when the regular internal username/password
	verification is used. The {@link #EXTERNAL} is a generic external verification but
	external {@link AuthenticationManager}:s are encouraged to define their own authentication
	methods.
	
	@author Nicklas
	@since 3.3
*/
public class AuthenticationMethod 
{

	private static Map<String, AuthenticationMethod> methods = new HashMap<String, AuthenticationMethod>();
	
	/**
		This is the regular internal username/password verification authentication method.
	*/
	public static AuthenticationMethod INTERNAL = AuthenticationMethod.getInstance("internal");
	
	/**
		This is a generic external authentication method. {@link AuthenticationManager}:s are 
		encouraged to define their own authentication methods.
	*/
	public static AuthenticationMethod EXTERNAL = AuthenticationMethod.getInstance("external");	

	/**
		Get an instance of this class for the given authentication method. If no
		method with that name exists a new method is created.
		@param method The name of the method
	*/
	public static synchronized final AuthenticationMethod getInstance(String method)
	{
		AuthenticationMethod am = methods.get(method);
		if (am == null)
		{
			am = new AuthenticationMethod(method);
			methods.put(method, am);
		}
		return am;
	}
	
	private final String method;
	
	private AuthenticationMethod(String method)
	{
		this.method = method;
	}
	
	/**
		Get the name of the authentication method.
	*/
	public String getMethod()
	{
		return method;
	}
	
	@Override
	public String toString()
	{
		return method;
	}
	
}
