/*
	$Id: EntityQuery.java 6897 2015-05-12 07:25:48Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import java.util.Collection;
import java.util.List;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Permission;

/**
	This a HQL query that return enitities, ie. items. The items may be returned
	in the form of data classes ({@link net.sf.basedb.core.DataQuery}) or
	in the form of item classes ({@link net.sf.basedb.core.ItemQuery}).
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-05-12 09:25:48 +0200 (ti, 12 maj 2015) $
*/
public interface EntityQuery
	extends HqlQuery
{

	/**
		The type of items the query returns.
	*/
	public Item getItemType();
	
	/**
		The type of items that is the root of the query. In most cases,
		this is the same as the {@link #getItemType()}, but not
		always.
		@since 2.8
	*/
	public Item getRootType();
	
	/**
		Specify options for which items to include in the result. 

		@param includes An array of items that should be included
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #exclude(Include[])
	*/
	public void include(Include... includes);
	
	/**
		@see #include(Include[])
	*/
	public void include(Collection<Include> includes);
	
	/**
	 	Clears the lists of includes and adds a new collection.
	 	@see #include(Include...)
	 	
	 	@param includes The new includes to be hold in the list.
	 	@since 2.14.2	 	
	 */
	public void setIncludes(Collection<Include> includes);
	
	/**
		Specify options for which items to exclude from the result. 

		@param excludes An array of options that should be excluded
		@throws UnsupportedOperationException If this operation isn't supported by
			the implementation
		@see #include(Include[])
	*/
	public void exclude(Include... excludes);
	
	/**
		@see #exclude(Include[])
	*/
	public void exclude(Collection<Include> excludes);
	
	/**
		Check which options are set for the items to include in the result
		@param includes The options to check
		@return TRUE If all specified options are included, FALSE otherwise
	*/
	public boolean isIncluded(Include... includes);
	
	/**
		@see #isIncluded(Include[])
	*/
	public boolean isIncluded(Collection<Include> includes);
	
	/**
		Only return items which the logged in user has the specified 
		permission for. The default is to return all items which the user 
		has at least read permission for. 
		@param permission The required permission
	*/
	public void setItemPermission(Permission permission);
	

	/**
	 	The logged in user's permission on the items returned by this query.
		@return A Permission object
	*/
	public Permission getItemPermission();


	/**
		Execute the query and return the ID values of all items matching
		the query.
		@param dc The <code>DbControl</code> used to access the database
			and check permissions
		@throws BaseException If there is an error
		@since 3.5
	*/
	public List<Integer> idList(DbControl dc);

}
