/**
	$Id: SectionAssaysParser.java 5096 2009-09-14 06:29:17Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.importer.spotdata;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.basefile.BaseFileParser;
import net.sf.basedb.util.basefile.BaseFileSectionParser;
import net.sf.basedb.util.importer.spotdata.BaseFileInfo.ChildBioAssay;
import net.sf.basedb.util.parser.FlatFileParser;

/**
	Parses the 'assays' section of a BASEfile with bioassay set spot
	data. This section is optional, but if it exists, it must define 
	'id' and 'name' columns of child bioassays. The 'id' must be the 
	same as the id of a parent bioassay unless a 'parents' column is 
	present. One or more parents can be given for each child. Separate
	the values with a /.
	<p>
	NOTE! The column names can be redefined by calling {@link 
	BaseFileParser#setRedefinedColumnName(String, String, String)}
	
	<pre class="code">
section	assays
columns	id\tname\tparents
count	2
%
100	The first child	1/2/3
101	The second child	4/5/6
</pre>

	@author Nicklas
	@version 2.14
	@base.modified $Date: 2009-09-14 08:29:17 +0200 (må, 14 sep 2009) $
*/
public class SectionAssaysParser
	implements BaseFileSectionParser
{

	private final DbControl dc;
	private final BaseFileInfo info;
	
	
	/**
		Creates a new 'section assays' parser.
		@param dc The DbControl to use for database access
		@param info Information collector for the file we are parsing
	*/
	public SectionAssaysParser(DbControl dc, BaseFileInfo info)
	{
		this.dc = dc;
		this.info = info;
	}
	
	/*
		From the BaseFileSectionParser interface
		-----------------------------------------
	*/
	@Override
	public void parseSection(BaseFileParser parser, FlatFileParser ffp)
		throws IOException
	{
		String filename = info.getName();
		
		// Parse the headers and verify that they meet the requirements
		ffp.parseHeaders();
		int line = ffp.getParsedLines();
		List<String> columns = parser.getRequiredHeader(ffp, "columns", "\\t", "assays", filename);
		int idIndex = parser.getRequiredIndex(columns, 
				parser.getRedefinedColumnName("assays", "id"), "columns", "assays", line, filename);
		int nameIndex = parser.getRequiredIndex(columns, 
				parser.getRedefinedColumnName("assays", "name"), "columns", "assays", line, filename);
		int parentIndex = columns.indexOf(parser.getRedefinedColumnName("assays", "parents"));
		
		// Store for future reference
		boolean hasParentAssaysMapping = parentIndex != -1;
		info.setAssaysHaveParentAssaysMapping(hasParentAssaysMapping);
		
		// Read data part of the 'assays' section
		// Each line has ID, NAME and PARENTS (optional)
		ffp.setMinDataColumns(columns.size());
		FlatFileParser.Data data;
		while ((data = ffp.nextData()) != null)
		{
			line = ffp.getParsedLines();
			
			int assayId = Values.getInt(data.get(idIndex));
			ChildBioAssay child = new ChildBioAssay();
			child.setFileId(assayId);
			child.setName(data.get(nameIndex));
			if (hasParentAssaysMapping)
			{
				Set<Integer> parents = new HashSet<Integer>(Arrays.asList(Values.getInt(data.get(parentIndex).split("/"))));
				Iterator<Integer> it = parents.iterator();
				while (it.hasNext())
				{
					// The given parent must already exist in the database
					info.checkBioAssay(dc, it.next(), line);
				}
				child.setParents(parents);
			}
			else
			{
				// If no explicit parent mapping is defined, the id of the child must 
				// reference the parent bioassay
				info.checkBioAssay(dc, assayId, line);
			}
			
			if (!info.addChildAssay(child))
			{
				throw new BaseException("Duplicate assay id '" + assayId + 
						"' found in section 'assays' at line " + line + 
						" in file '" + filename + "'");
			}
		}
	}
	// -------------------------------------------
	
	
}
