/**
	$Id: SpotIntensityParser.java 5096 2009-09-14 06:29:17Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.importer.spotdata;

import java.util.List;

import net.sf.basedb.util.basefile.BaseFileParser;
import net.sf.basedb.util.parser.FlatFileParser;

/**
	A spot intensity parser is responsible for parsing and extracting
	spot intensity data on a row-by-row basis from a BASEfile. 
	After parsing the headers of a 'spot' section each registered
	spot intensity parser will be asked if the section contains
	all needed data columns {@link #hasRequiredAssayFields(BaseFileParser, FlatFileParser, List)}.
	The first parser that gives a positive response will be used to 
	parse the spot intensity data for that section.
	<p>
	The actual parsing is done in the second pass.

	@see FirstPassSectionSpotsParser#addSpotIntensityParser(SpotIntensityParser)
	@author Nicklas
	@version 2.14
	@base.modified $Date: 2009-09-14 08:29:17 +0200 (må, 14 sep 2009) $
*/
public interface SpotIntensityParser
{
	/**
		Check the list of assay fields to see if all columns that are
		required are present or not. This method is called in the 
		first pass to decide which spot intensity parser that should
		be used in the second pass.
		
		@param parser The master BASEfile parser
		@param ffp The flat file parser that parses the current file
		@param assayFields The list of assay fields
		@return TRUE if all required fields are present, FALSE otherwise
	*/
	public boolean hasRequiredAssayFields(BaseFileParser parser, FlatFileParser ffp, List<String> assayFields);
	
	/**
		Prepare for parsing the data section. Preparations typically include
		extracting the column indexes of the data columns. This method is 
		called in the second pass before starting to parse data for
		a 'spot' section. NOTE! The same object instance may be used to
		parse more than one section in a serial BASEfile.

		@param parser The master BASEfile parser
		@param ffp The flat file parser that parses the current file
		@param assayFields The list of assay fields
	*/
	public void beginSection(BaseFileParser parser, FlatFileParser ffp, List<String> assayFields);
	
	/**
		Extract the intensity values from the current data line for a single 
		bioassay. If the section contains data for multiple bioassays
		(eg. a matrix BASEfile) this method is called multiple times (with
		different 'firstIndex' value) for each row.
		
		@param data The current data line
		@param intensities An array that should be populate with intensity values
		@param firstIndex The index of the first data column that contains data
			for the current bioassay
			
		@return TRUE if the values could be set correctly, FALSE otherwise (which
			means that the spot data will not be inserted)
	*/
	public boolean setIntensities(FlatFileParser.Data data, float[] intensities, int firstIndex);
	
	
}
