/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.Date;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioPlateEvent;
import net.sf.basedb.core.DateUtil;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Kit;
import net.sf.basedb.core.MeasuredBioMaterial;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.Subtypable;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for kits. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_KIT}
	<li>Access denied: {@link Validator#DENIED_KIT}
	<li>Incorrect hardware type: {@link Validator#INCORRECT_KITTYPE} 
	<li>Using kit after expiration date: {@link Validator#KIT_USED_AFTER_EXPIRATIONDATE}
	</ul>

	@author Nicklas
	@since 3.7
*/
public class KitValidator
	extends NameableNodeValidator<Kit>
{
	
	public KitValidator()
	{
		super(null, Validator.DENIED_KIT);
	}

	
	/* 
		From BasicValidator class
		-------------------------
	*/
	/**
		If the parent item has a subtype that is related to a KIT subtype, report
		the missing item as a {@link Validator#MISSING_KIT} failure. Otherwise,
		ignore the missing kit.
		@return Always false
	*/
	@Override
	public boolean preMissingItem(DbControl dc, OverviewContext context, Node parentNode)
	{
		BasicItem parentItem = parentNode.getItem(dc);
		ItemSubtype expectedSubtype = getExpectedKitSubtype(dc, parentItem, false);
		if (expectedSubtype != null)
		{
			context.createFailure(Validator.MISSING_KIT, parentNode, 
					"Missing kit: " + expectedSubtype.getName(), getMissingItemFix(dc, parentNode));
		}
		return false;
	}
	
	/**
		Checks if the hardware is of the correct hardware type
		and, if a project is active, the hardware is included as a
		default hardware for the project.
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		Kit kit = (Kit)node.getItem(dc);
		
		BasicItem parentItem = parentNode.getItem(dc);
		ItemSubtype expectedSubtype = getExpectedKitSubtype(dc, parentItem, true);
		try
		{
			ItemSubtype subtype = kit.getItemSubtype();
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				ItemSubtype mySubtype = ((Subtypable)parentItem).getItemSubtype();
				Fix subtypeFix = subtype == null || mySubtype == null ? null : 
					new Fix("Change related kit subtype of '" + mySubtype.getName() + 
						"' to '" + subtype.getName() + "'", mySubtype);

				context.createFailure(Validator.INCORRECT_KITTYPE, parentNode, 
					"Expected a kit with subtype: " + expectedSubtype.getName(),
					new Fix("Change kit of '" + parentNode.getTitle() + "'", parentItem),
					new Fix("Change kit type of '" + kit.getName() + "' to '" + expectedSubtype.getName() + "'", kit),
					subtypeFix
				);
			}
		}
		catch (PermissionDeniedException ex)
		{}
		
		// Check expiration date
		Date expirationDate = kit.getExpirationDate();
		if (expirationDate != null)
		{
			Date useDate = null;
			if (parentItem instanceof MeasuredBioMaterial)
			{
				useDate = ((MeasuredBioMaterial)parentItem).getCreationEvent().getEventDate();
			}
			else if (parentItem instanceof PhysicalBioAssay)
			{
				useDate = ((PhysicalBioAssay)parentItem).getCreationEvent().getEventDate();
			}
			else if (parentItem instanceof DerivedBioAssay)
			{
				// There is no 'event date' on this item so we can't check this
			}
			else if (parentItem instanceof BioPlateEvent)
			{
				useDate = ((BioPlateEvent)parentItem).getEventDate();
			}
			
			if (useDate != null && useDate.after(expirationDate))
			{
				context.createFailure(Validator.KIT_USED_AFTER_EXPIRATIONDATE, 
					parentNode, "Kit used after expiration date: " + DateUtil.formatDate(expirationDate), 
					new Fix("Change event date of '" + parentNode.getTitle() + "'", parentItem),
					new Fix("Change expiration date of kit '" + kit.getName() + "'", kit)
				);
			}
		}
	}
	
	/**
		@return Suggested fix is to add a hardware to the parent item
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		BasicItem parentItem = parentNode.getItem();		
		return new Fix("Add kit to '" + parentNode.getTitle() + "'", parentItem);
	}
	// --------------------------------
	
	
	/**
		Get the subtype of the associated kit that we expect for the given
		parent item.
	*/
	protected ItemSubtype getExpectedKitSubtype(DbControl dc, BasicItem parentItem, boolean useDefault)
	{
		ItemSubtype expectedSubtype = null;
		int defaultSubtypeForItemType = SystemItems.getId(Kit.getDefaultSystemId(parentItem.getType()));
		try
		{
			if (parentItem instanceof Subtypable)
			{
				expectedSubtype = ItemSubtype.getRelatedSubtype(dc, 
						(Subtypable)parentItem, Item.KIT, useDefault ? defaultSubtypeForItemType : 0);
			}
			else if (useDefault && defaultSubtypeForItemType != 0)
			{
				expectedSubtype = ItemSubtype.getById(dc, defaultSubtypeForItemType);
			}
		}
		catch (PermissionDeniedException ex)
		{}
		return expectedSubtype;
	}

	
}
