/**
	$Id: TarFileUnpacker.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PushbackInputStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.zip.GZIPInputStream;

import org.apache.tools.bzip2.CBZip2InputStream;

import com.ice.tar.TarEntry;
import com.ice.tar.TarInputStream;

import net.sf.basedb.core.AbsoluteProgressReporter;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.File;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.Location;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.InputStreamTracker;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.zip.AbstractFileUnpacker;
import net.sf.basedb.util.zip.PackUtil;
import net.sf.basedb.util.zip.TarUtil;

/**
	A plugin for unpacking tar, gzipped tar and bzipped tar files to
	the BASE file system.
	
	@author nicklas
	@version 2.0
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fr, 14 sep 2012) $
*/
public class TarFileUnpacker
	extends AbstractFileUnpacker
{

	private static final Set<String> extensions =  
		new LinkedHashSet<String>(Arrays.asList("tar", "tar.gz", "tgz", "tar.bz2"));

	private static final Set<String> mimeTypes =  
		new LinkedHashSet<String>(Arrays.asList("application/x-tar", "application/x-gzip", 
			"application/x-bzip2"));

	private Map<String, Directory> paths;
	
	/*
		From the FileUnpacker interface
		-------------------------------------------
	*/
	@Override
	public String getFormatName()
	{
		return "Tar files";
	}
	/**
		Return a set containing the strings [ tar, tar.gz, tgz, tar.bz2 ].
	*/
	@Override
	public Set<String> getExtensions()
	{
		return extensions;
	}
	/**
		Return a set containing the strings [ application/x-tar, application/x-gzip, 
		application/x-bzip2 ].
	*/
	@Override
	public Set<String> getMimeTypes()
	{
		return mimeTypes;
	}
	@Override
	public int unpack(DbControl dc, Directory rootDir, InputStream in,
		File sourceFile, boolean overwrite, AbsoluteProgressReporter progress)
		throws IOException
	{
		InputStreamTracker tracker = new InputStreamTracker(in);
		PushbackInputStream pin = new PushbackInputStream(tracker, 10);
		TarInputStream tarStream = null;
		
		// Check if the file is a 'gzip' of 'bzip' file. If not we assume it is pure 'tar'.
		boolean isGzip = FileUtil.checkMagicNumber(pin, new byte[] { 0x1f, (byte)0x8b } );
		boolean isBzip = !isGzip && 
			FileUtil.checkMagicNumber(pin, new byte[] { 0x42, 0x5a, 0x68, 0x39, 0x31 } );
		if (isGzip)
		{
			tarStream = new TarInputStream(new GZIPInputStream(new BufferedInputStream(pin)));
		}
		else if (isBzip)
		{
			// Strip first two bytes 'BZ', since they are not part of the file format
			pin.read();
			pin.read();
			tarStream = new TarInputStream(new CBZip2InputStream(new BufferedInputStream(pin)));
		}
		else
		{
			tarStream = new TarInputStream(pin);
		}
		
		this.paths = new HashMap<String, Directory>();
		
		TarEntry entry = null;
		String rootPath = rootDir.getPath().toString();
		paths.put(rootPath, rootDir);
		
		long totalUnpacked = 0;
		int numFiles = 0;
		while ((entry = TarUtil.getNextEntry(tarStream)) != null)
		{
			ThreadSignalHandler.checkInterrupted();
			String subPath = entry.getName();
			String completePath = rootPath + "/" + subPath;
			
			boolean isDirectory = entry.isDirectory();
			if (progress != null)
			{
				totalUnpacked = tracker.getNumRead();
				progress.displayAbsolute(totalUnpacked, 
					"Unpacking " + (isDirectory ? "directory: " : "file: ") + completePath + 
					" (" +Values.formatBytes(entry.getSize()) + ")");
			}
			if (!isDirectory)
			{
				Path filePath = new Path(completePath, Path.Type.FILE);

				// Get the parent directory
				Directory parentDir = getDirectory(dc, filePath);
				
				// Get the file or create a new one
				File file = File.getFile(dc, parentDir, filePath.getFilename(), true);
				if (!file.isInDatabase() || file.getLocation() == Location.OFFLINE || overwrite)
				{
					if (!file.isInDatabase()) dc.saveItem(file);
					file.setMimeTypeAuto(null, null);
					OutputStream upload = file.getUploadStream(false);
					PackUtil.copyProperties(sourceFile, file);
					try
					{
						FileUtil.copy(tarStream, upload);
						numFiles++;
					}
					finally
					{
						upload.close();
					}
				}
			}
		}
		if (progress != null) progress.display(100, "Done");
		return numFiles;
	}
	// -------------------------------------------

	/**
		Get the directory for the specified path. We first check the internal cache.
		If not found we try to load it from the database. If still not found we create
		a new directory and save it to the database.
		@param path The path
		@return The directory
	*/
	private Directory getDirectory(DbControl dc, Path path)
	{
		// Make sure it is a path to the directory
		if (path.getType() == Path.Type.FILE) path = path.getParent();
		String thePath = path.toString();
		
		// Load directory from cache
		Directory dir = paths.get(thePath);
		
		if (dir == null)
		{
			// Load directory from database
			try
			{
				dir = Directory.getByPath(dc, path);
				dir.setRemoved(false);
			}
			catch (ItemNotFoundException ex)
			{}
		}
		if (dir == null)
		{
			// Directory is not in the cache and not in the database --> create a new one
			// Get the parent recursively
			Directory parent = getDirectory(dc, path.getParent());
			dir = Directory.getNew(dc, parent);
			dc.saveItem(dir);
			dir.setName(path.getDirectory(path.getDirectoryCount()-1));
		}
		paths.put(thePath, dir);
		return dir;	
	}
	
}
