/*
	$Id: TestBioSource.java 5340 2010-05-10 06:15:00Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;
import net.sf.basedb.util.FileUtil;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.security.cert.X509Certificate;
import java.util.Set;

public class TestFileServer
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing file servers");
		write_header();
		// Standard tests: create, load, list
		int id = test_create("base.thep.lu.se", "BASE trac site", "base", "base", null, null, null);
		int id2 = test_create("base.onk.lu.se", "BASE demo site", null, null, "data/base.onk.lu.se.crt", null, null);
		test_load(id);
		test_list(-1);

		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		// Standard test: Delete
		test_delete(id);
		test_delete(id2);
		write("++Testing file servers "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(String name, String description, String username, String password, 
		String serverCertificateFile, String clientCertificateFile, String clientCertificatePassword)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.FILESERVER)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			FileServer fs = FileServer.getNew(dc);
			fs.setName(name);
			fs.setDescription(description);
			fs.setUsername(username);
			fs.setPassword(password);
			if (serverCertificateFile != null)
			{
				ByteArrayOutputStream out = new ByteArrayOutputStream();
				FileUtil.copy(new FileInputStream(serverCertificateFile), out);
				out.close();
				fs.setServerCertificate(out.toByteArray());
			}
			if (clientCertificateFile != null)
			{
				ByteArrayOutputStream out = new ByteArrayOutputStream();
				FileUtil.copy(new FileInputStream(clientCertificateFile), out);
				out.close();
				fs.setClientCertificate(out.toByteArray());
				fs.setClientCertificatePassword(clientCertificatePassword);
			}
			dc.saveItem(fs);
			dc.commit();
			id = fs.getId();
			write_item(0, fs);
			write("--Create file server OK");
		}
		catch (Throwable ex)
		{
			write("--Create file server FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			FileServer fs = FileServer.getById(dc, id);
			write_item(0, fs);
			write("--Load file server OK");
		}
		catch (Throwable ex)
		{
			write("--Load file server FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<FileServer> l = FileServer.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List file server OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List file server FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			FileServer fs = FileServer.getById(dc, id);
			dc.deleteItem(fs);
			Set<ItemProxy> using = fs.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(fs + " is used by " + using);
			}
			dc.commit();
			write("--Delete file server OK");
		}
		catch (Throwable ex)
		{
			write("--Delete file server FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}


	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tUsername\tServer certificate");
			write("-- \t-- \t--------- \t-----------\t--------\t------------------");
		}
	}
	static void write_item(int i, FileServer fs)
		throws BaseException
	{
		
		if (!TestUtil.getSilent()) 
		{
			String cert = null;
			try
			{
				X509Certificate tmp = fs.getServerX509Certificate();
				cert = tmp.getSubjectX500Principal().toString();
			}
			catch (Exception ex)
			{
				cert = ex.getMessage();
			}
			System.out.println(i+":\t"+fs.getId()+"\t"+fs.getName()+"\t"+fs.getDescription()+
					"\t"+fs.getUsername() + "\t" + cert);
		}
	}
	static void write(String message)
	{
		System.out.println(message);
	}
}
